=== modified file 'README.txt'
--- README.txt	2006-12-11 06:44:19 +0000
+++ README.txt	2006-12-19 09:46:12 +0000
@@ -1,4 +1,56 @@
-loggerhead
-
-This is a TurboGears (http://www.turbogears.org) project. It can be
-started by running the start-loggerhead.py script.
\ No newline at end of file
+LOGGERHEAD
+==========
+
+Loggerhead is a web viewer for bazaar branches.  It can be used to navigate
+a branch history, annotate files, perform searches... all the usual things.
+
+This is a TurboGears (http://www.turbogears.org) project.  It can be
+started by running the 'start-loggerhead.py' script.
+
+If you're not familiar with turbogears, the easiest way to get this script
+started is to run it via::
+
+    $ nohup ./start-loggerhead.py &
+    
+and add the following lines to your apache configuration::
+
+    <Location "/bzr.dev/">
+        ProxyPass http://127.0.0.1:8080/
+        ProxyPassReverse http://127.0.0.1:8080/
+    </Location>
+    
+The port configuration is in "dev.cfg".
+
+The config file is "loggerhead.conf".  Currently it can only serve one
+branch at a time, but in the future I plan to add support for serving
+multiple branches at once.
+
+
+CACHES
+------
+
+To speed up most operations, loggerhead will start creating two caches
+when it first launches:
+
+    - a revision data cache
+    - a text searching cache
+
+You can put the cache folder anywhere, but I find that a folder under
+the branch's .bzr/ folder is the simplest place.
+
+The caches for a branch with 10,000 revisions take about 15 minutes each
+on my machine, but YMMV.  Once they are built, they update every six hours
+or so but usually finish quickly (or instantly).  Until the revision cache
+is built, all operations will be slow.
+
+You may update the bazaar branch at any time (for example, from a cron).
+Loggerhead will notice and refresh, and bazaar uses its own branch locking
+to prevent corruption.
+
+
+SUPPORT
+-------
+
+This is loosely based on bazaar-webserve, which was loosely based on hgweb.
+Discussion should take place on the bazaar-dev mailing list.
+

=== modified file 'dev.cfg'
--- dev.cfg	2006-12-15 09:20:04 +0000
+++ dev.cfg	2006-12-19 02:19:42 +0000
@@ -55,12 +55,12 @@
 [[loggers]]
 [[[loggerhead]]]
 level='DEBUG'
-qualname='loggerhead'
+#qualname='loggerhead'
 handlers=['debug_out']
 
-[[[allinfo]]]
-level='INFO'
-handlers=['debug_out']
+#[[[allinfo]]]
+#level='INFO'
+#handlers=['debug_out']
 
 [[[access]]]
 level='INFO'

=== modified file 'loggerhead.conf'
--- loggerhead.conf	2006-12-15 09:20:04 +0000
+++ loggerhead.conf	2006-12-20 09:29:42 +0000
@@ -1,16 +1,25 @@
 
-# where is the bazaar branch located, and what should we call it?
-folder='/Users/robey/code/bzr/bzr.dev'
-branch_name='bzr.dev'
-
-# should we cache revisions?  (HIGHLY recommended)  where?
-# this will be a folder containing a couple of files.  if the folder
-# doesn't already exist, we'll create it.
-cachepath='/Users/robey/code/bzr/bzr.dev/.bzr/loggerhead-cache'
-
-# what url should i use in the atom feed, for external sites?
-external_url="http://localhost:8080"
-
 # where's bzrlib?
+# (you only need this if you don't install bzr.)
 bzrpath='/Users/robey/code/bzr/bzr'
 
+# each branch is in a sub-folder of the website, and the config
+# key describes that folder name:
+[bazaar-dev]
+
+    # where is the bazaar branch located, and what should we call it?
+    folder='/Users/robey/code/bzr/bzr.dev'
+
+    # (optional) is there a friendier name for the branch that i should
+    # use on the web pages?
+    #branch_name='bazaar-dev'
+
+    # should we cache revisions?  (HIGHLY recommended)  where?
+    # this will be a folder containing a couple of files.  if the folder
+    # doesn't already exist, we'll create it.
+    cachepath='/Users/robey/code/bzr/bzr.dev/.bzr/loggerhead-cache'
+
+    # what is my externally-visible url? (for use in the atom feed)
+    external_url="http://localhost:8080"
+
+

=== modified file 'loggerhead/TODO'
--- loggerhead/TODO	2006-12-15 09:20:04 +0000
+++ loggerhead/TODO	2006-12-20 09:29:42 +0000
@@ -1,19 +1,18 @@
 
-* config file
-
-* maybe put this at the footer of each page:
-    paramiko-dev, 419 revisions, 1133 days old
-    first revision: Tue 2003-11-04 08:34:24 +0000
-    latest revision: Sat 2006-12-02 18:33:34 -0800
-    ...or somewhere.
-    ... does this really matter?
-
-* add back the search bar.
-    -- should notice dotted-revnos and use that to jump.
-	-- also search by date should be simple.
-
 * problem: given a folder, find all revisions where items inside that
   folder have changed.
     -- stumper.  maybe ask on bazaar-dev ML?
 
-	
\ No newline at end of file
+* side by side diff?  matty says "overflow-x" css property for horiz scrollbar
+
+* make equivalent of start page:
+    http://goffredo-baroncelli.homelinux.net/bazaar
+
+* on revision page: view/download patch
+
+* potentially do a 'bzr pull' on a branch on an interval timer
+
+* give this thing a version # or something
+* make a loggerhead webpage
+
+* primary / page listing branches

=== added file 'loggerhead/branchview.py'
--- loggerhead/branchview.py	1970-01-01 00:00:00 +0000
+++ loggerhead/branchview.py	2006-12-20 09:29:42 +0000
@@ -0,0 +1,103 @@
+#
+# Copyright (C) 2006  Robey Pointer <robey@lag.net>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+"""
+collection of configuration and objects related to a bazaar branch.
+"""
+
+import logging
+import threading
+
+import turbogears
+from cherrypy import HTTPRedirect
+
+from loggerhead import util
+from loggerhead.changecache import ChangeCache
+from loggerhead.history import History
+from loggerhead.textindex import TextIndex
+from loggerhead.controllers.changelog_ui import ChangeLogUI
+from loggerhead.controllers.atom_ui import AtomUI
+from loggerhead.controllers.revision_ui import RevisionUI
+from loggerhead.controllers.inventory_ui import InventoryUI
+from loggerhead.controllers.annotate_ui import AnnotateUI
+from loggerhead.controllers.download_ui import DownloadUI
+
+
+with_history_lock = util.with_lock('_history_lock')
+
+
+class BranchView (object):
+    def __init__(self, name, config):
+        self._name = name
+        self._config = config
+        self.log = logging.getLogger('loggerhead.%s' % (name,))
+        
+        # branch history
+        self._history_lock = threading.RLock()
+        self._history = None
+        
+        self.changes = ChangeLogUI(self)
+        self.revision = RevisionUI(self)
+        self.files = InventoryUI(self)
+        self.annotate = AnnotateUI(self)
+        self.download = DownloadUI(self)
+        self.atom = AtomUI(self)
+        
+        # force history object to be loaded:
+        self.get_history()
+    
+    config = property(lambda self: self._config)
+    
+    name = property(lambda self: self._name)
+    
+    friendly_name = property(lambda self: self._config.get('branch_name', self._name))
+
+    @turbogears.expose()
+    def index(self):
+        raise HTTPRedirect(self.url('/changes'))
+
+    @with_history_lock
+    def get_history(self):
+        """
+        get an up-to-date History object, safely.  each page-view calls this
+        method, and normally it will get the same History object as on previous
+        calls.  but if the bazaar branch on-disk has been updated since this
+        History was created, a new object will be created and returned.
+        """
+        if (self._history is None) or self._history.out_of_date():
+            self.log.debug('Reload branch history...')
+            if self._history is not None:
+                self._history.detach()
+            self._history = History.from_folder(self._config.get('folder'), self._name)
+            cache_path = self._config.get('cachepath', None)
+            if cache_path is not None:
+                self._history.use_cache(ChangeCache(self._history, cache_path))
+                self._history.use_search_index(TextIndex(self._history, cache_path))
+        return self._history
+    
+    def check_rebuild(self):
+        h = self.get_history()
+        h.check_rebuild()
+    
+    def url(self, elements, **kw):
+        if not isinstance(elements, list):
+            elements = [elements]
+        if elements[0].startswith('/'):
+            elements[0] = elements[0][1:]
+        return turbogears.url([ '/' + self.name ] + elements, **kw)
+

=== added file 'loggerhead/changecache.py'
--- loggerhead/changecache.py	1970-01-01 00:00:00 +0000
+++ loggerhead/changecache.py	2006-12-20 09:29:42 +0000
@@ -0,0 +1,158 @@
+#
+# Copyright (C) 2006  Robey Pointer <robey@lag.net>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+"""
+a cache for chewed-up "change" data structures, which are basically just a
+different way of storing a revision.  the cache improves lookup times 10x
+over bazaar's xml revision structure, though, so currently still worth doing.
+
+once a revision is committed in bazaar, it never changes, so once we have
+cached a change, it's good forever.
+"""
+
+import logging
+import os
+import shelve
+import threading
+import time
+
+from loggerhead import util
+from loggerhead.util import decorator
+
+
+with_lock = util.with_lock('_lock')
+
+
+class ChangeCache (object):
+    
+    def __init__(self, history, cache_path):
+        self.history = history
+        self.log = history.log
+        
+        if not os.path.exists(cache_path):
+            os.mkdir(cache_path)
+
+        # keep a separate cache for the diffs, because they're very time-consuming to fetch.
+        changes_filename = os.path.join(cache_path, 'changes')
+        changes_diffs_filename = os.path.join(cache_path, 'changes-diffs')
+        
+        self._cache = shelve.open(changes_filename, 'c', protocol=2)
+        self._cache_diffs = shelve.open(changes_diffs_filename, 'c', protocol=2)
+        
+        self._lock = threading.RLock()
+        self._closed = False
+        
+        # once we process a change (revision), it should be the same forever.
+        self.log.info('Using change cache %s; %d/%d entries.' % (cache_path, len(self._cache), len(self._cache_diffs)))
+    
+    @with_lock
+    def close(self):
+        self._cache.close()
+        self._cache_diffs.close()
+        self._closed = True
+    
+    @with_lock
+    def closed(self):
+        return self._closed
+
+    @with_lock
+    def flush(self):
+        self._cache.sync()
+        self._cache_diffs.sync()
+    
+    @with_lock
+    def get_changes(self, revid_list, get_diffs=False):
+        """
+        get a list of changes by their revision_ids.  any changes missing
+        from the cache are fetched by calling L{History.get_change_uncached}
+        and inserted into the cache before returning.
+        """
+        if get_diffs:
+            cache = self._cache_diffs
+        else:
+            cache = self._cache
+
+        out = []
+        fetch_list = []
+        sfetch_list = []
+        for revid in revid_list:
+            # if the revid is in unicode, use the utf-8 encoding as the key
+            srevid = util.to_utf8(revid)
+            
+            if srevid in cache:
+                out.append(cache[srevid])
+            else:
+                #self.log.debug('Entry cache miss: %r' % (revid,))
+                out.append(None)
+                fetch_list.append(revid)
+                sfetch_list.append(srevid)
+        
+        if len(fetch_list) > 0:
+            # some revisions weren't in the cache; fetch them
+            changes = self.history.get_changes_uncached(fetch_list, get_diffs)
+            if changes is None:
+                return changes
+            for i in xrange(len(revid_list)):
+                if out[i] is None:
+                    cache[sfetch_list.pop(0)] = out[i] = changes.pop(0)
+        
+        return out
+    
+    @with_lock
+    def full(self, get_diffs=False):
+        if get_diffs:
+            cache = self._cache_diffs
+        else:
+            cache = self._cache
+        return (len(cache) >= len(self.history.get_revision_history())) and (util.to_utf8(self.history.last_revid) in cache)
+
+    def check_rebuild(self, max_time=3600):
+        """
+        check if we need to fill in any missing pieces of the cache.  pull in
+        any missing changes, but don't work any longer than C{max_time}
+        seconds.
+        """
+        if self.full():
+            return
+        
+        self.log.info('Building revision cache...')
+        start_time = time.time()
+        last_update = time.time()
+        count = 0
+
+        work = list(self.history.get_revision_history())
+        jump = 100
+        for i in xrange(0, len(work), jump):
+            r = work[i:i + jump]
+            self.get_changes(r)
+            if self.closed():
+                return
+            count += jump
+            now = time.time()
+            if now - start_time > max_time:
+                self.log.info('Cache rebuilding will pause for now.')
+                self.flush()
+                return
+            if now - last_update > 60:
+                self.log.info('Revision cache rebuilding continues: %d/%d' % (min(count, len(work)), len(work)))
+                last_update = time.time()
+                self.flush()
+        self.log.info('Revision cache rebuild completed.')
+        self.flush()
+
+

=== modified file 'loggerhead/controllers/__init__.py'
--- loggerhead/controllers/__init__.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/__init__.py	2006-12-20 09:29:42 +0000
@@ -18,6 +18,7 @@
 #
 
 import logging
+import sys
 import time
 
 from configobj import ConfigObj
@@ -26,62 +27,48 @@
 from turbogears import controllers
 from cherrypy import HTTPRedirect, NotFound
 
+my_config = ConfigObj('loggerhead.conf', encoding='utf-8')
+extra_path = my_config.get('bzrpath', None)
+if extra_path:
+    sys.path.insert(0, extra_path)
+
 from loggerhead import util
-
-my_config = ConfigObj('loggerhead.conf', encoding='utf-8')
-util.set_config(my_config)
-
-from loggerhead.controllers.changelog_ui import ChangeLogUI
-from loggerhead.controllers.atom_ui import AtomUI
-from loggerhead.controllers.revision_ui import RevisionUI
-from loggerhead.controllers.inventory_ui import InventoryUI
-from loggerhead.controllers.annotate_ui import AnnotateUI
+from loggerhead.branchview import BranchView
 from loggerhead.history import History
 
 log = logging.getLogger("loggerhead.controllers")
 
+
 class Root (controllers.RootController):
-    changes = ChangeLogUI()
-    atom = AtomUI()
-    revision = RevisionUI()
-    files = InventoryUI()
-    annotate = AnnotateUI()
-    
+    def __init__(self):
+        global my_config
+        self._views = []
+        for branch_name in my_config.sections:
+            log.debug('Configuring branch %r...', branch_name)
+            view = BranchView(branch_name, my_config[branch_name])
+            setattr(self, branch_name, view)
+            self._views.append(view)
+        
     @turbogears.expose()
     def index(self):
-        raise HTTPRedirect(turbogears.url('/changes'))
-
-
-# force history to be read:
-util.get_history()
-
-
-
-def rebuild_cache():
-    log.info('Rebuilding revision cache...')
-    last_update = time.time()
-    h = util.get_history()
-    count = 0
-    
-    work = h.get_revision_history()
-    for r in work:
-        h.get_change(r)
-        if h.out_of_date():
-            return
-        count += 1
-        if time.time() - last_update > 60:
-            log.info('Revision cache rebuilding continues: %d/%d' % (count, len(work)))
-            last_update = time.time()
-            h.flush_cache()
-    log.info('Revision cache rebuild completed.')
-
-
-# re-index every hour (for now -- maybe should be even longer?)
-index_freq = 3600
-
-turbogears.scheduler.add_interval_task(initialdelay=1, interval=index_freq, action=rebuild_cache)
+        # FIXME - display list of branches
+        raise HTTPRedirect(turbogears.url('/bazaar-dev/changes'))
+
+    def check_rebuild(self):
+        for v in self._views:
+            v.check_rebuild()
+
+
+# singleton:
+Root = Root()
+
+# re-index every 6 hours
+index_freq = 6 * 3600
+
+turbogears.scheduler.add_interval_task(initialdelay=1, interval=index_freq, action=Root.check_rebuild)
 
 # for use in profiling the very-slow get_change() method:
 #h = util.get_history()
-#h._get_change_profiled(h.last_revid)
+#w = list(h.get_revision_history())
+#h._get_changes_profiled(w[:100])
 

=== modified file 'loggerhead/controllers/annotate_ui.py'
--- loggerhead/controllers/annotate_ui.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/annotate_ui.py	2006-12-20 09:29:42 +0000
@@ -41,40 +41,50 @@
         
 class AnnotateUI (object):
 
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+
     @turbogears.expose(html='loggerhead.templates.annotate')
     def default(self, *args, **kw):
         z = time.time()
-        h = util.get_history()
+        h = self._branch.get_history()
         
         if len(args) > 0:
             revid = h.fix_revid(args[0])
         else:
             revid = None
         
-        path = kw.get('path', None)
-        if (path == '/') or (path == '') or (path is None):
-            raise HTTPRedirect(turbogears.url('/changes'))
+        file_id = kw.get('file_id', None)
+        if file_id is None:
+            raise HTTPRedirect(self._branch.url('/changes'))
 
         try:
-            revlist, revid = h.get_navigation(revid, path)
-            file_id = h.get_inventory(revid).path2id(path)
+            revid_list, revid = h.get_file_view(revid, file_id)
         except Exception, x:
-            log.error('Exception fetching changes: %r, %s' % (x, x))
-            raise HTTPRedirect(turbogears.url('/changes'))
+            self.log.error('Exception fetching changes: %s' % (x,))
+            util.log_exception(self.log)
+            raise HTTPRedirect(self._branch.url('/changes'))
             
         # no navbar for revisions
         navigation = util.Container()
+        
+        path = h.get_path(revid, file_id)
+        filename = os.path.basename(path)
 
         vals = {
-            'branch_name': util.get_config().get('branch_name'),
+            'branch': self._branch,
             'util': util,
             'revid': revid,
+            'file_id': file_id,
             'path': path,
+            'filename': filename,
             'history': h,
             'navigation': navigation,
-            'change': h.get_change(revid),
+            'change': h.get_changes([ revid ])[0],
             'contents': list(h.annotate_file(file_id, revid)),
         }
         h.flush_cache()
-        log.info('/annotate: %r secs' % (time.time() - z,))
+        self.log.info('/annotate: %r secs' % (time.time() - z,))
         return vals

=== modified file 'loggerhead/controllers/atom_ui.py'
--- loggerhead/controllers/atom_ui.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/atom_ui.py	2006-12-20 09:29:42 +0000
@@ -24,21 +24,22 @@
 
 class AtomUI (object):
     
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+
     @turbogears.expose(template='loggerhead.templates.atom', format="xml", content_type="application/atom+xml")
     def default(self, *args):
-        h = util.get_history()
+        h = self._branch.get_history()
+        pagesize = int(self._branch.config.get('pagesize', '20'))
 
-        # really, there is no point giving a revid here...
-        if len(args) > 0:
-            revid = args[0]
-        else:
-            revid = h.last_revid
-        revlist = h.get_short_revision_history_from(revid)
-        entries = list(h.get_changelist(list(revlist)[:20]))
+        revid_list, start_revid = h.get_file_view(None, None)
+        entries = list(h.get_changes(list(revid_list)[:pagesize]))
 
         vals = {
-            'external_url': util.get_config().get('external_url'),
-            'branch_name': util.get_config().get('branch_name'),
+            'external_url': self._branch.config.get('external_url'),
+            'branch': self._branch,
             'changes': entries,
             'util': util,
             'history': h,

=== modified file 'loggerhead/controllers/changelog_ui.py'
--- loggerhead/controllers/changelog_ui.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/changelog_ui.py	2006-12-20 09:29:42 +0000
@@ -17,10 +17,7 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
 
-import base64
-import logging
 import os
-import posixpath
 import time
 
 import turbogears
@@ -28,8 +25,6 @@
 
 from loggerhead import util
 
-log = logging.getLogger("loggerhead.controllers")
-
 
 # just thinking out loud here...
 #
@@ -38,68 +33,79 @@
 #         the current beginning of navigation (navigation continues back to
 #         the original revision) -- this may not be along the primary revision
 #         path since the user may have navigated into a branch
-#     - path
+#     - file_id
 #         if navigating the revisions that touched a file
 #     - current revid
 #         current location along the navigation path (while browsing)
 #
-# current revid is given on the url path.  'path' and 'starting revid' are
+# current revid is given on the url path.  'file_id' and 'starting revid' are
 # handed along as params.
 
 
 class ChangeLogUI (object):
     
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+        
     @turbogears.expose(html='loggerhead.templates.changelog')
     def default(self, *args, **kw):
         z = time.time()
-        h = util.get_history()
+        h = self._branch.get_history()
+        config = self._branch.config
         
         if len(args) > 0:
             revid = h.fix_revid(args[0])
         else:
             revid = None
 
-        path = kw.get('path', None)
+        file_id = kw.get('file_id', None)
+        query = kw.get('q', None)
         start_revid = h.fix_revid(kw.get('start_revid', None))
-        pagesize = int(util.get_config().get('pagesize', '20'))
+        orig_start_revid = start_revid
+        pagesize = int(config.get('pagesize', '20'))
+        search_failed = False
         
         try:
-            revlist, start_revid = h.get_navigation(start_revid, path)
-            if revid is None:
-                revid = start_revid
-            if revid not in revlist:
-                # if the given revid is not in the revlist, use a revlist that
-                # starts at the given revid.
-                revlist, start_revid = h.get_navigation(revid, path)
-            entry_list = list(h.get_revids_from(revlist, revid))[:pagesize]
-            entries = h.get_changelist(entry_list)
+            revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
+            if (query is not None) and (len(revid_list) == 0):
+                search_failed = True
+
+            if len(revid_list) == 0:
+                scan_list = revid_list
+            else:
+                scan_list = revid_list[h.get_revid_sequence(revid_list, revid):]
+            entry_list = scan_list[:pagesize]
+            entries = h.get_changes(entry_list)
         except Exception, x:
-            log.error('Exception fetching changes: %r, %s' % (x, x))
-            raise HTTPRedirect(turbogears.url('/changes'))
+            self.log.error('Exception fetching changes: %s' % (x,))
+            util.log_exception(self.log)
+            raise HTTPRedirect(self._branch.url('/changes'))
 
-        navigation = util.Container(pagesize=pagesize, revid=revid, start_revid=start_revid, revlist=revlist,
-                                    path=path, scan_url='/changes', feed=True)
+        navigation = util.Container(pagesize=pagesize, revid=revid, start_revid=start_revid, revid_list=revid_list,
+                                    file_id=file_id, scan_url='/changes', feed=True)
+        if query is not None:
+            navigation.query = query
         util.fill_in_navigation(h, navigation)
         
         entries = list(entries)
         # add parent & merge-point branch-nick info, in case it's useful
-        for change in entries:
-            for p in change.parents:
-                p.branch_nick = h.get_change(p.revid).branch_nick
-            for p in change.merge_points:
-                p.branch_nick = h.get_change(p.revid).branch_nick
+        h.get_branch_nicks(entries)
 
         vals = {
-            'branch_name': util.get_config().get('branch_name'),
+            'branch': self._branch,
             'changes': list(entries),
             'util': util,
             'history': h,
             'revid': revid,
             'navigation': navigation,
-            'path': path,
-            'last_revid': h.last_revid,
+            'file_id': file_id,
             'start_revid': start_revid,
+            'viewing_from': (orig_start_revid is not None) and (orig_start_revid != h.last_revid),
+            'query': query,
+            'search_failed': search_failed,
         }
         h.flush_cache()
-        log.info('/changes %r: %r secs' % (revid, time.time() - z))
+        self.log.info('/changes %r: %r secs' % (revid, time.time() - z))
         return vals

=== added file 'loggerhead/controllers/download_ui.py'
--- loggerhead/controllers/download_ui.py	1970-01-01 00:00:00 +0000
+++ loggerhead/controllers/download_ui.py	2006-12-20 09:29:42 +0000
@@ -0,0 +1,63 @@
+#
+# Copyright (C) 2006  Robey Pointer <robey@lag.net>
+# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+import datetime
+import logging
+import mimetypes
+import os
+import posixpath
+import textwrap
+import time
+
+import turbogears
+from cherrypy import HTTPRedirect, response, session
+
+from loggerhead import util
+
+
+log = logging.getLogger("loggerhead.controllers")
+
+
+class DownloadUI (object):
+
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+
+    @turbogears.expose()
+    def default(self, *args, **kw):
+        # /download/<rev_id>/<file_id>/[filename]
+        z = time.time()
+        h = self._branch.get_history()
+        
+        if len(args) < 2:
+            raise HTTPRedirect(self._branch.url('/changes'))
+        
+        revid = h.fix_revid(args[0])
+        file_id = args[1]
+        filename, content = h.get_file(file_id, revid)
+        mime_type, encoding = mimetypes.guess_type(filename)
+        if mime_type is None:
+            mime_type = 'application/octet-stream'
+        
+        response.headers['Content-Type'] = mime_type
+        response.headers['Content-Length'] = len(content)
+        response.body = content
+        return response.body

=== modified file 'loggerhead/controllers/inventory_ui.py'
--- loggerhead/controllers/inventory_ui.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/inventory_ui.py	2006-12-20 09:29:42 +0000
@@ -41,56 +41,67 @@
         
 class InventoryUI (object):
 
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+
     @turbogears.expose(html='loggerhead.templates.inventory')
     def default(self, *args, **kw):
         z = time.time()
-        h = util.get_history()
+        h = self._branch.get_history()
         
         if len(args) > 0:
             revid = h.fix_revid(args[0])
         else:
             revid = None
         
-        path = kw.get('path', None)
-        if (path == '/') or (path == ''):
-            path = None
+        file_id = kw.get('file_id', None)
+        sort_type = kw.get('sort', None)
 
         try:
-            revlist, revid = h.get_navigation(revid, path)
+            revid_list, revid = h.get_file_view(revid, file_id)
             rev = h.get_revision(revid)
             inv = h.get_inventory(revid)
         except Exception, x:
-            log.error('Exception fetching changes: %r, %s' % (x, x))
-            raise HTTPRedirect(turbogears.url('/changes'))
+            self.log.error('Exception fetching changes: %s' % (x,))
+            util.log_exception(self.log)
+            raise HTTPRedirect(self._branch.url('/changes'))
 
-        if path is None:
-            path = '/'
-            file_id = None
-        else:
-            file_id = inv.path2id(path)
-            
         # no navbar for revisions
         navigation = util.Container()
 
+        change = h.get_changes([ revid ])[0]
         # add parent & merge-point branch-nick info, in case it's useful
-        change = h.get_change(revid)
-        for p in change.parents:
-            p.branch_nick = h.get_change(p.revid).branch_nick
-        for p in change.merge_points:
-            p.branch_nick = h.get_change(p.revid).branch_nick
+        h.get_branch_nicks([ change ])
+        
+        path = inv.id2path(file_id)
+        if not path.startswith('/'):
+            path = '/' + path
+        idpath = inv.get_idpath(file_id)
+        if len(idpath) > 1:
+            updir = dirname(path)
+            updir_file_id = idpath[-2]
+        else:
+            updir = None
+            updir_file_id = None
+        if updir == '/':
+            updir_file_id = None
 
         vals = {
-            'branch_name': util.get_config().get('branch_name'),
+            'branch': self._branch,
             'util': util,
             'revid': revid,
             'change': change,
+            'file_id': file_id,
             'path': path,
-            'updir': dirname(path),
-            'filelist': h.get_filelist(inv, path),
+            'updir': updir,
+            'updir_file_id': updir_file_id,
+            'filelist': h.get_filelist(inv, path, sort_type),
             'history': h,
             'posixpath': posixpath,
             'navigation': navigation,
         }
         h.flush_cache()
-        log.info('/inventory %r: %r secs' % (revid, time.time() - z))
+        self.log.info('/inventory %r: %r secs' % (revid, time.time() - z))
         return vals

=== modified file 'loggerhead/controllers/revision_ui.py'
--- loggerhead/controllers/revision_ui.py	2006-12-15 09:20:04 +0000
+++ loggerhead/controllers/revision_ui.py	2006-12-20 09:29:42 +0000
@@ -33,49 +33,53 @@
 
 class RevisionUI (object):
 
+    def __init__(self, branch):
+        # BranchView object
+        self._branch = branch
+        self.log = branch.log
+
     @turbogears.expose(html='loggerhead.templates.revision')
     def default(self, *args, **kw):
         z = time.time()
-        h = util.get_history()
+        h = self._branch.get_history()
         
         if len(args) > 0:
             revid = h.fix_revid(args[0])
         else:
             revid = None
         
-        path = kw.get('path', None)
+        file_id = kw.get('file_id', None)
         start_revid = h.fix_revid(kw.get('start_revid', None))
+        query = kw.get('q', None)
         
         try:
-            revlist, start_revid = h.get_navigation(start_revid, path)
-            if revid is None:
-                revid = start_revid
+            revid, start_revid, revid_list = h.get_view(revid, start_revid, file_id, query)
         except Exception, x:
-            log.error('Exception fetching changes: %r, %s' % (x, x))
-            raise HTTPRedirect(turbogears.url('/changes'))
+            self.log.error('Exception fetching changes: %s' % (x,))
+            util.log_exception(self.log)
+            raise HTTPRedirect(self._branch.url('/changes'))
         
-        navigation = util.Container(revlist=revlist, revid=revid, start_revid=start_revid, path=path,
-                                    pagesize=1, scan_url='/revision', feed=1)
+        navigation = util.Container(revid_list=revid_list, revid=revid, start_revid=start_revid, file_id=file_id,
+                                    pagesize=1, scan_url='/revision', feed=True)
+        if query is not None:
+            navigation.query = query
         util.fill_in_navigation(h, navigation)
 
-        change = h.get_change(revid, get_diffs=True)
-        
+        change = h.get_changes([ revid ], get_diffs=True)[0]
         # add parent & merge-point branch-nick info, in case it's useful
-        for p in change.parents:
-            p.branch_nick = h.get_change(p.revid).branch_nick
-        for p in change.merge_points:
-            p.branch_nick = h.get_change(p.revid).branch_nick
+        h.get_branch_nicks([ change ])
         
         vals = {
-            'branch_name': util.get_config().get('branch_name'),
+            'branch': self._branch,
             'revid': revid,
             'change': change,
             'start_revid': start_revid,
-            'path': path,
+            'file_id': file_id,
             'util': util,
             'history': h,
             'navigation': navigation,
+            'query': query,
         }
         h.flush_cache()
-        log.info('/revision: %r seconds' % (time.time() - z,))
+        self.log.info('/revision: %r seconds' % (time.time() - z,))
         return vals

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2006-12-15 09:20:04 +0000
+++ loggerhead/history.py	2006-12-20 09:29:42 +0000
@@ -1,6 +1,8 @@
 #
 # Copyright (C) 2006  Robey Pointer <robey@lag.net>
 # Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
+# Copyright (C) 2005  Jake Edge <jake@edge2.net>
+# Copyright (C) 2005  Matt Mackall <mpm@selenic.com>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,6 +19,14 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
 
+#
+# This file (and many of the web templates) contains work based on the
+# "bazaar-webserve" project by Goffredo Baroncelli, which is in turn based
+# on "hgweb" by Jake Edge and Matt Mackall.
+#
+
+
+import bisect
 import cgi
 import datetime
 import logging
@@ -31,9 +41,7 @@
 from StringIO import StringIO
 
 from loggerhead import util
-extra_path = util.get_config().get('bzrpath', None)
-if extra_path:
-    sys.path.insert(0, extra_path)
+from loggerhead.util import decorator
 
 import bzrlib
 import bzrlib.annotate
@@ -46,22 +54,7 @@
 import bzrlib.ui
 
 
-log = logging.getLogger("loggerhead.controllers")
-
-
-# cache lock binds tighter than branch lock
-def with_cache_lock(unbound):
-    def cache_locked(self, *args, **kw):
-        self._cache_lock.acquire()
-        try:
-            return unbound(self, *args, **kw)
-        finally:
-            self._cache_lock.release()
-    cache_locked.__doc__ = unbound.__doc__
-    cache_locked.__name__ = unbound.__name__
-    return cache_locked
-
-
+@decorator
 def with_branch_lock(unbound):
     def branch_locked(self, *args, **kw):
         self._lock.acquire()
@@ -69,11 +62,20 @@
             return unbound(self, *args, **kw)
         finally:
             self._lock.release()
-    branch_locked.__doc__ = unbound.__doc__
-    branch_locked.__name__ = unbound.__name__
     return branch_locked
 
 
+@decorator
+def with_bzrlib_read_lock(unbound):
+    def bzrlib_read_locked(self, *args, **kw):
+        self._branch.repository.lock_read()
+        try:
+            return unbound(self, *args, **kw)
+        finally:
+            self._branch.repository.unlock()
+    return bzrlib_read_locked
+
+
 # bzrlib's UIFactory is not thread-safe
 uihack = threading.local()
 
@@ -86,28 +88,44 @@
 bzrlib.ui.ui_factory = ThreadSafeUIFactory()
 
 
+# from bzrlib
+class _RevListToTimestamps(object):
+    """This takes a list of revisions, and allows you to bisect by date"""
+
+    __slots__ = ['revid_list', 'repository']
+
+    def __init__(self, revid_list, repository):
+        self.revid_list = revid_list
+        self.repository = repository
+
+    def __getitem__(self, index):
+        """Get the date of the index'd item"""
+        return datetime.datetime.fromtimestamp(self.repository.get_revision(self.revid_list[index]).timestamp)
+
+    def __len__(self):
+        return len(self.revid_list)
+
+
 class History (object):
     
     def __init__(self):
         self._change_cache = None
-        self._cache_lock = threading.Lock()
+        self._index = None
         self._lock = threading.RLock()
     
-    def __del__(self):
-        if self._change_cache is not None:
-            self._change_cache.close()
-            self._change_cache_diffs.close()
-            self._change_cache = None
-            self._change_cache_diffs = None
-
     @classmethod
-    def from_branch(cls, branch):
+    def from_branch(cls, branch, name=None):
         z = time.time()
         self = cls()
         self._branch = branch
         self._history = branch.revision_history()
-        self._revision_graph = branch.repository.get_revision_graph()
         self._last_revid = self._history[-1]
+        self._revision_graph = branch.repository.get_revision_graph(self._last_revid)
+        
+        if name is None:
+            name = self._branch.nick
+        self._name = name
+        self.log = logging.getLogger('loggerhead.%s' % (name,))
         
         self._full_history = []
         self._revision_info = {}
@@ -130,13 +148,13 @@
             for parent in self._revision_graph[revid]:
                 self._where_merged.setdefault(parent, set()).add(revid)
 
-        log.info('built revision graph cache: %r secs' % (time.time() - z,))
+        self.log.info('built revision graph cache: %r secs' % (time.time() - z,))
         return self
     
     @classmethod
-    def from_folder(cls, path):
+    def from_folder(cls, path, name=None):
         b = bzrlib.branch.Branch.open(path)
-        return cls.from_branch(b)
+        return cls.from_branch(b, name)
 
     @with_branch_lock
     def out_of_date(self):
@@ -144,46 +162,35 @@
             return True
         return False
 
-    @with_cache_lock
-    def use_cache(self, path):
-        if not os.path.exists(path):
-            os.mkdir(path)
-        # keep a separate cache for the diffs, because they're very time-consuming to fetch.
-        cachefile = os.path.join(path, 'changes')
-        cachefile_diffs = os.path.join(path, 'changes-diffs')
-        
-        # why can't shelve allow 'cw'?
-        if not os.path.exists(cachefile):
-            self._change_cache = shelve.open(cachefile, 'c', protocol=2)
-        else:
-            self._change_cache = shelve.open(cachefile, 'w', protocol=2)
-        if not os.path.exists(cachefile_diffs):
-            self._change_cache_diffs = shelve.open(cachefile_diffs, 'c', protocol=2)
-        else:
-            self._change_cache_diffs = shelve.open(cachefile_diffs, 'w', protocol=2)
-            
-        # once we process a change (revision), it should be the same forever.
-        log.info('Using change cache %s; %d, %d entries.' % (path, len(self._change_cache), len(self._change_cache_diffs)))
-        self._change_cache_filename = cachefile
-        self._change_cache_diffs_filename = cachefile_diffs
-
-    @with_cache_lock
-    def dont_use_cache(self):
-        # called when a new history object needs to be created.  we can't use
-        # the cache files anymore; they belong to the new history object.
-        if self._change_cache is None:
-            return
-        self._change_cache.close()
-        self._change_cache_diffs.close()
-        self._change_cache = None
-        self._change_cache_diffs = None
-
-    @with_cache_lock
+    def use_cache(self, cache):
+        self._change_cache = cache
+    
+    def use_search_index(self, index):
+        self._index = index
+
+    @with_branch_lock
+    def detach(self):
+        # called when a new history object needs to be created, because the
+        # branch history has changed.  we need to immediately close and stop
+        # using our caches, because a new history object will be created to
+        # replace us, using the same cache files.
+        if self._change_cache is not None:
+            self._change_cache.close()
+            self._change_cache = None
+        if self._index is not None:
+            self._index.close()
+            self._index = None
+
     def flush_cache(self):
         if self._change_cache is None:
             return
-        self._change_cache.sync()
-        self._change_cache_diffs.sync()
+        self._change_cache.flush()
+    
+    def check_rebuild(self):
+        if self._change_cache is not None:
+            self._change_cache.check_rebuild()
+        if self._index is not None:
+            self._index.check_rebuild()
     
     last_revid = property(lambda self: self._last_revid, None, None)
     
@@ -233,7 +240,7 @@
             if len(parents) == 0:
                 return
             revid = parents[0]
-
+    
     @with_branch_lock
     def get_short_revision_history_by_fileid(self, file_id):
         # wow.  is this really the only way we can get this list?  by
@@ -245,7 +252,97 @@
         revids = [r for r in self._full_history if r in w_revids]
         return revids
 
+    @with_branch_lock
+    def get_revision_history_since(self, revid_list, date):
+        # if a user asks for revisions starting at 01-sep, they mean inclusive,
+        # so start at midnight on 02-sep.
+        date = date + datetime.timedelta(days=1)
+        # our revid list is sorted in REVERSE date order, so go thru some hoops here...
+        revid_list.reverse()
+        index = bisect.bisect(_RevListToTimestamps(revid_list, self._branch.repository), date)
+        if index == 0:
+            return []
+        revid_list.reverse()
+        index = -index
+        return revid_list[index:]
+    
+    @with_branch_lock
+    def get_revision_history_matching(self, revid_list, text):
+        self.log.debug('searching %d revisions for %r', len(revid_list), text)
+        z = time.time()
+        # this is going to be painfully slow. :(
+        out = []
+        text = text.lower()
+        for revid in revid_list:
+            change = self.get_changes([ revid ])[0]
+            if text in change.comment.lower():
+                out.append(revid)
+        self.log.debug('searched %d revisions for %r in %r secs', len(revid_list), text, time.time() - z)
+        return out
+
+    def get_revision_history_matching_indexed(self, revid_list, text):
+        self.log.debug('searching %d revisions for %r', len(revid_list), text)
+        z = time.time()
+        if self._index is None:
+            return self.get_revision_history_matching(revid_list, text)
+        out = self._index.find(text, revid_list)
+        self.log.debug('searched %d revisions for %r in %r secs: %d results', len(revid_list), text, time.time() - z, len(out))
+        # put them in some coherent order :)
+        out = [r for r in self._full_history if r in out]
+        return out
+    
+    @with_branch_lock
+    def get_search_revid_list(self, query, revid_list):
+        """
+        given a "quick-search" query, try a few obvious possible meanings:
+        
+            - revision id or # ("128.1.3")
+            - date (US style "mm/dd/yy", earth style "dd-mm-yy", or iso style "yyyy-mm-dd")
+            - comment text as a fallback
+
+        and return a revid list that matches.
+        """
+        # FIXME: there is some silliness in this action.  we have to look up
+        # all the relevant changes (time-consuming) only to return a list of
+        # revids which will be used to fetch a set of changes again.
+        
+        # if they entered a revid, just jump straight there; ignore the passed-in revid_list
+        revid = self.fix_revid(query)
+        if revid is not None:
+            changes = self.get_changes([ revid ])
+            if (changes is not None) and (len(changes) > 0):
+                return [ revid ]
+        
+        date = None
+        m = self.us_date_re.match(query)
+        if m is not None:
+            date = datetime.datetime(util.fix_year(int(m.group(3))), int(m.group(1)), int(m.group(2)))
+        else:
+            m = self.earth_date_re.match(query)
+            if m is not None:
+                date = datetime.datetime(util.fix_year(int(m.group(3))), int(m.group(2)), int(m.group(1)))
+            else:
+                m = self.iso_date_re.match(query)
+                if m is not None:
+                    date = datetime.datetime(util.fix_year(int(m.group(1))), int(m.group(2)), int(m.group(3)))
+        if date is not None:
+            if revid_list is None:
+                # if no limit to the query was given, search only the direct-parent path.
+                revid_list = list(self.get_revids_from(None, self._last_revid))
+            return self.get_revision_history_since(revid_list, date)
+        
+        # check comment fields.
+        if revid_list is None:
+            revid_list = self._full_history
+        return self.get_revision_history_matching_indexed(revid_list, query)
+    
     revno_re = re.compile(r'^[\d\.]+$')
+    # the date regex are without a final '$' so that queries like
+    # "2006-11-30 12:15" still mostly work.  (i think it's better to give
+    # them 90% of what they want instead of nothing at all.)
+    us_date_re = re.compile(r'^(\d{1,2})/(\d{1,2})/(\d\d(\d\d?))')
+    earth_date_re = re.compile(r'^(\d{1,2})-(\d{1,2})-(\d\d(\d\d?))')
+    iso_date_re = re.compile(r'^(\d\d\d\d)-(\d\d)-(\d\d)')
 
     def fix_revid(self, revid):
         # if a "revid" is actually a dotted revno, convert it to a revid
@@ -256,31 +353,89 @@
         return revid
     
     @with_branch_lock
-    def get_navigation(self, revid, path):
+    def get_file_view(self, revid, file_id):
         """
         Given an optional revid and optional path, return a (revlist, revid)
         for navigation through the current scope: from the revid (or the
         latest revision) back to the original revision.
         
-        If path is None, the entire revision history is the list scope.
+        If file_id is None, the entire revision history is the list scope.
         If revid is None, the latest revision is used.
         """
         if revid is None:
             revid = self._last_revid
-        if path is not None:
+        if file_id is not None:
             # since revid is 'start_revid', possibly should start the path tracing from revid... FIXME
             inv = self._branch.repository.get_revision_inventory(revid)
-            revlist = list(self.get_short_revision_history_by_fileid(inv.path2id(path)))
+            revlist = list(self.get_short_revision_history_by_fileid(file_id))
+            revlist = list(self.get_revids_from(revlist, revid))
         else:
             revlist = list(self.get_revids_from(None, revid))
         if revid is None:
             revid = revlist[0]
         return revlist, revid
+    
+    @with_branch_lock
+    def get_view(self, revid, start_revid, file_id, query=None):
+        """
+        use the URL parameters (revid, start_revid, file_id, and query) to
+        determine the revision list we're viewing (start_revid, file_id, query)
+        and where we are in it (revid).
+        
+        if a query is given, we're viewing query results.
+        if a file_id is given, we're viewing revisions for a specific file.
+        if a start_revid is given, we're viewing the branch from a
+            specific revision up the tree.
+        (these may be combined to view revisions for a specific file, from
+            a specific revision, with a specific search query.)
+            
+        returns a new (revid, start_revid, revid_list, scan_list) where:
+        
+            - revid: current position within the view
+            - start_revid: starting revision of this view
+            - revid_list: list of revision ids for this view
+        
+        file_id and query are never changed so aren't returned, but they may
+        contain vital context for future url navigation.
+        """
+        if query is None:
+            revid_list, start_revid = self.get_file_view(start_revid, file_id)
+            if revid is None:
+                revid = start_revid
+            if revid not in revid_list:
+                # if the given revid is not in the revlist, use a revlist that
+                # starts at the given revid.
+                revid_list, start_revid = self.get_file_view(revid, file_id)
+            return revid, start_revid, revid_list
+        
+        # potentially limit the search
+        if (start_revid is not None) or (file_id is not None):
+            revid_list, start_revid = self.get_file_view(start_revid, file_id)
+        else:
+            revid_list = None
+
+        revid_list = self.get_search_revid_list(query, revid_list)
+        if len(revid_list) > 0:
+            if revid not in revid_list:
+                revid = revid_list[0]
+            return revid, start_revid, revid_list
+        else:
+            # no results
+            return None, None, []
 
     @with_branch_lock
     def get_inventory(self, revid):
         return self._branch.repository.get_revision_inventory(revid)
 
+    @with_branch_lock
+    def get_path(self, revid, file_id):
+        if (file_id is None) or (file_id == ''):
+            return ''
+        path = self._branch.repository.get_revision_inventory(revid).id2path(file_id)
+        if (len(path) > 0) and not path.startswith('/'):
+            path = '/' + path
+        return path
+    
     def get_where_merged(self, revid):
         try:
             return self._where_merged[revid]
@@ -333,166 +488,137 @@
                 d[revnos] = ( revnolast, revid )
 
         return [ d[revnos][1] for revnos in d.keys() ]
-            
-    def get_changelist(self, revid_list):
-        for revid in revid_list:
-            yield self.get_change(revid)
+
+    def get_branch_nicks(self, changes):
+        """
+        given a list of changes from L{get_changes}, fill in the branch nicks
+        on all parents and merge points.
+        """
+        fetch_set = set()
+        for change in changes:
+            for p in change.parents:
+                fetch_set.add(p.revid)
+            for p in change.merge_points:
+                fetch_set.add(p.revid)
+        p_changes = self.get_changes(list(fetch_set))
+        p_change_dict = dict([(c.revid, c) for c in p_changes])
+        for change in changes:
+            for p in change.parents:
+                p.branch_nick = p_change_dict[p.revid].branch_nick
+            for p in change.merge_points:
+                p.branch_nick = p_change_dict[p.revid].branch_nick
     
     @with_branch_lock
-    def get_change(self, revid, get_diffs=False):
+    def get_changes(self, revid_list, get_diffs=False):
         if self._change_cache is None:
-            return self._get_change(revid, get_diffs)
-
-        # if the revid is in unicode, use the utf-8 encoding as the key
-        srevid = revid
-        if isinstance(revid, unicode):
-            srevid = revid.encode('utf-8')
-        return self._get_change_from_cache(revid, srevid, get_diffs)
-
-    @with_cache_lock
-    def _get_change_from_cache(self, revid, srevid, get_diffs):
-        if get_diffs:
-            cache = self._change_cache_diffs
-        else:
-            cache = self._change_cache
-            
-        if srevid in cache:
-            c = cache[srevid]
-        else:
-            if get_diffs and (srevid in self._change_cache):
-                # salvage the non-diff entry for a jump-start
-                c = self._change_cache[srevid]
-                if len(c.parents) == 0:
-                    left_parent = None
-                else:
-                    left_parent = c.parents[0].revid
-                c.changes = self.diff_revisions(revid, left_parent, get_diffs=True)
-                cache[srevid] = c
-            else:
-                #log.debug('Entry cache miss: %r' % (revid,))
-                c = self._get_change(revid, get_diffs=get_diffs)
-                cache[srevid] = c
-            
+            changes = self.get_changes_uncached(revid_list, get_diffs)
+        else:
+            changes = self._change_cache.get_changes(revid_list, get_diffs)
+        if changes is None:
+            return changes
+        
         # some data needs to be recalculated each time, because it may
         # change as new revisions are added.
-        merge_revids = self.simplify_merge_point_list(self.get_merge_point_list(revid))
-        c.merge_points = [util.Container(revid=r, revno=self.get_revno(r)) for r in merge_revids]
+        for i in xrange(len(revid_list)):
+            revid = revid_list[i]
+            change = changes[i]
+            merge_revids = self.simplify_merge_point_list(self.get_merge_point_list(revid))
+            change.merge_points = [util.Container(revid=r, revno=self.get_revno(r)) for r in merge_revids]
         
-        return c
-    
+        return changes
+
     # alright, let's profile this sucka.
-    def _get_change_profiled(self, revid, get_diffs=False):
+    def _get_changes_profiled(self, revid_list, get_diffs=False):
         from loggerhead.lsprof import profile
         import cPickle
-        ret, stats = profile(self._get_change, revid, get_diffs)
+        ret, stats = profile(self.get_changes_uncached, revid_list, get_diffs)
         stats.sort()
         stats.freeze()
         cPickle.dump(stats, open('lsprof.stats', 'w'), 2)
         return ret
 
-    def _get_change(self, revid, get_diffs=False):
+    @with_bzrlib_read_lock
+    def get_changes_uncached(self, revid_list, get_diffs=False):
         try:
-            rev = self._branch.repository.get_revision(revid)
+            rev_list = self._branch.repository.get_revisions(revid_list)
         except (KeyError, bzrlib.errors.NoSuchRevision):
-            # ghosted parent?
+            return None
+        
+        delta_list = self._branch.repository.get_deltas_for_revisions(rev_list)
+        combined_list = zip(rev_list, delta_list)
+        
+        tree_map = {}
+        if get_diffs:
+            # lookup the trees for each revision, so we can calculate diffs
+            lookup_set = set()
+            for rev in rev_list:
+                lookup_set.add(rev.revision_id)
+                if len(rev.parent_ids) > 0:
+                    lookup_set.add(rev.parent_ids[0])
+            tree_map = dict((t.get_revision_id(), t) for t in self._branch.repository.revision_trees(lookup_set))
+            # also the root tree, in case we hit the origin:
+            tree_map[None] = self._branch.repository.revision_tree(None)
+        
+        entries = []
+        for rev, delta in combined_list:
+            commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
+            
+            parents = [util.Container(revid=r, revno=self.get_revno(r)) for r in rev.parent_ids]
+    
+            if len(parents) == 0:
+                left_parent = None
+            else:
+                left_parent = rev.parent_ids[0]
+            
+            message = rev.message.splitlines()
+            if len(message) == 1:
+                # robey-style 1-line long message
+                message = textwrap.wrap(message[0])
+            
+            # make short form of commit message
+            short_message = message[0]
+            if len(short_message) > 60:
+                short_message = short_message[:60] + '...'
+    
+            old_tree, new_tree = None, None
+            if get_diffs:
+                new_tree = tree_map[rev.revision_id]
+                old_tree = tree_map[left_parent]
+
             entry = {
-                'revid': 'missing',
-                'revno': '',
-                'date': datetime.datetime.fromtimestamp(0),
-                'author': 'missing',
-                'branch_nick': None,
-                'short_comment': 'missing',
-                'comment': 'missing',
-                'comment_clean': 'missing',
-                'parents': [],
-                'merge_points': [],
-                'changes': [],
+                'revid': rev.revision_id,
+                'revno': self.get_revno(rev.revision_id),
+                'date': commit_time,
+                'author': rev.committer,
+                'branch_nick': rev.properties.get('branch-nick', None),
+                'short_comment': short_message,
+                'comment': rev.message,
+                'comment_clean': [util.html_clean(s) for s in message],
+                'parents': parents,
+                'changes': self.parse_delta(delta, get_diffs, old_tree, new_tree),
             }
-            log.error('ghost entry: %r' % (revid,))
-            return util.Container(entry)
-            
-        commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
-        
-        parents = [util.Container(revid=r, revno=self.get_revno(r)) for r in rev.parent_ids]
-
-        if len(parents) == 0:
-            left_parent = None
-        else:
-            left_parent = rev.parent_ids[0]
-        
-        message = rev.message.splitlines()
-        if len(message) == 1:
-            # robey-style 1-line long message
-            message = textwrap.wrap(message[0])
-        
-        # make short form of commit message
-        short_message = message[0]
-        if len(short_message) > 60:
-            short_message = short_message[:60] + '...'
-
-        entry = {
-            'revid': revid,
-            'revno': self.get_revno(revid),
-            'date': commit_time,
-            'author': rev.committer,
-            'branch_nick': rev.properties.get('branch-nick', None),
-            'short_comment': short_message,
-            'comment': rev.message,
-            'comment_clean': [util.html_clean(s) for s in message],
-            'parents': parents,
-            'changes': self.diff_revisions(revid, left_parent, get_diffs=get_diffs),
-        }
-        return util.Container(entry)
-    
-    def scan_range(self, revlist, revid, pagesize=20):
-        """
-        yield a list of (label, title, revid) for a scan range through the full
-        branch history, centered around the given revid.
-        
-        example: [ ('<<', 'Previous page', 'rrr'), ('-10', 'Forward 10', 'rrr'),
-                   ('*', None, None), ('+10', 'Back 10', 'rrr'),
-                   ('+30', 'Back 30', 'rrr'), ('>>', 'Next page', 'rrr') ]
-        
-        next/prev page are always using the pagesize.
-        """
-        count = len(revlist)
-        pos = self.get_revid_sequence(revlist, revid)
-
-        if pos > 0:
-            yield (u'\xab', 'Previous page', revlist[max(0, pos - pagesize)])
-        else:
-            yield (u'\xab', None, None)
-        
-        offset_sign = -1
-        for offset in util.scan_range(pos, count, pagesize):
-            if (offset > 0) and (offset_sign < 0):
-                offset_sign = 0
-                # show current position
-#                yield ('[%s]' % (self.get_revno(revlist[pos]),), None, None)
-#                yield (u'\u2022', None, None)
-                yield (u'\u00b7', None, None)
-            if offset < 0:
-                title = 'Back %d' % (-offset,)
-            else:
-                title = 'Forward %d' % (offset,)
-            yield ('%+d' % (offset,), title, revlist[pos + offset])
-        
-        if pos < count - 1:
-            yield (u'\xbb', 'Next page', revlist[min(count - 1, pos + pagesize)])
-        else:
-            yield (u'\xbb', None, None)
-    
-    @with_branch_lock
-    def diff_revisions(self, revid, otherrevid, get_diffs=True):
-        """
-        Return a nested data structure containing the changes between two
-        revisions::
-        
-            added: list(filename),
-            renamed: list((old_filename, new_filename)),
-            deleted: list(filename),
+            entries.append(util.Container(entry))
+        
+        return entries
+
+    @with_branch_lock
+    def get_file(self, file_id, revid):
+        "returns (filename, data)"
+        inv_entry = self.get_inventory(revid)[file_id]
+        rev_tree = self._branch.repository.revision_tree(inv_entry.revision)
+        return inv_entry.name, rev_tree.get_file_text(file_id)
+    
+    @with_branch_lock
+    def parse_delta(self, delta, get_diffs=True, old_tree=None, new_tree=None):
+        """
+        Return a nested data structure containing the changes in a delta::
+        
+            added: list((filename, file_id)),
+            renamed: list((old_filename, new_filename, file_id)),
+            deleted: list((filename, file_id)),
             modified: list(
                 filename: str,
+                file_id: str,
                 chunks: list(
                     diff: list(
                         old_lineno: int,
@@ -505,11 +631,6 @@
         
         if C{get_diffs} is false, the C{chunks} will be omitted.
         """
-
-        new_tree = self._branch.repository.revision_tree(revid)
-        old_tree = self._branch.repository.revision_tree(otherrevid)
-        delta = new_tree.changes_from(old_tree)
-        
         added = []
         modified = []
         renamed = []
@@ -522,12 +643,6 @@
                 path += '@'
             return path
         
-        def tree_lines(tree, fid):
-            if not fid in tree:
-                return []
-            tree_file = bzrlib.textfile.text_file(tree.get_file(fid))
-            return tree_file.readlines()
-        
         def process_diff(diff):
             chunks = []
             chunk = None
@@ -567,33 +682,33 @@
                     
         def handle_modify(old_path, new_path, fid, kind):
             if not get_diffs:
-                modified.append(util.Container(filename=rich_filename(new_path, kind)))
+                modified.append(util.Container(filename=rich_filename(new_path, kind), file_id=fid))
                 return
-            old_lines = tree_lines(old_tree, fid)
-            new_lines = tree_lines(new_tree, fid)
+            old_lines = old_tree.get_file_lines(fid)
+            new_lines = new_tree.get_file_lines(fid)
             buffer = StringIO()
             bzrlib.diff.internal_diff(old_path, old_lines, new_path, new_lines, buffer)
             diff = buffer.getvalue()
-            modified.append(util.Container(filename=rich_filename(new_path, kind), chunks=process_diff(diff)))
+            modified.append(util.Container(filename=rich_filename(new_path, kind), file_id=fid, chunks=process_diff(diff), raw_diff=diff))
 
         for path, fid, kind in delta.added:
-            added.append(rich_filename(path, kind))
+            added.append((rich_filename(path, kind), fid))
         
         for path, fid, kind, text_modified, meta_modified in delta.modified:
             handle_modify(path, path, fid, kind)
         
         for oldpath, newpath, fid, kind, text_modified, meta_modified in delta.renamed:
-            renamed.append((rich_filename(oldpath, kind), rich_filename(newpath, kind)))
+            renamed.append((rich_filename(oldpath, kind), rich_filename(newpath, kind), fid))
             if meta_modified or text_modified:
                 handle_modify(oldpath, newpath, fid, kind)
         
         for path, fid, kind in delta.removed:
-            removed.append(rich_filename(path, kind))
+            removed.append((rich_filename(path, kind), fid))
         
         return util.Container(added=added, renamed=renamed, removed=removed, modified=modified)
 
     @with_branch_lock
-    def get_filelist(self, inv, path):
+    def get_filelist(self, inv, path, sort_type=None):
         """
         return the list of all files (and their attributes) within a given
         path subtree.
@@ -602,8 +717,16 @@
             path = path[:-1]
         if path.startswith('/'):
             path = path[1:]
-        parity = 0
-        for filepath, entry in inv.entries():
+        
+        entries = inv.entries()
+        
+        fetch_set = set()
+        for filepath, entry in entries:
+            fetch_set.add(entry.revision)
+        change_dict = dict([(c.revid, c) for c in self.get_changes(list(fetch_set))])
+        
+        file_list = []
+        for filepath, entry in entries:
             if posixpath.dirname(filepath) != path:
                 continue
             filename = posixpath.basename(filepath)
@@ -614,12 +737,28 @@
             
             # last change:
             revid = entry.revision
-            change = self.get_change(revid)
+            change = change_dict[revid]
             
-            yield util.Container(filename=filename, rich_filename=rich_filename, executable=entry.executable, kind=entry.kind,
-                                 pathname=pathname, revid=revid, change=change, parity=parity)
+            file = util.Container(filename=filename, rich_filename=rich_filename, executable=entry.executable, kind=entry.kind,
+                                  pathname=pathname, file_id=entry.file_id, size=entry.text_size, revid=revid, change=change)
+            file_list.append(file)
+        
+        if sort_type == 'filename':
+            file_list.sort(key=lambda x: x.filename)
+        elif sort_type == 'size':
+            file_list.sort(key=lambda x: x.size)
+        elif sort_type == 'date':
+            file_list.sort(key=lambda x: x.change.date)
+        
+        parity = 0
+        for file in file_list:
+            file.parity = parity
             parity ^= 1
-        pass
+
+        return file_list
+
+
+    _BADCHARS_RE = re.compile(ur'[\x00-\x08\x0b-\x0c\x0e-\x1f]')
 
     @with_branch_lock
     def annotate_file(self, file_id, revid):
@@ -629,12 +768,23 @@
         
         file_revid = self.get_inventory(revid)[file_id].revision
         oldvalues = None
-        revision_cache = {}
         
         # because we cache revision metadata ourselves, it's actually much
         # faster to call 'annotate_iter' on the weave directly than it is to
         # ask bzrlib to annotate for us.
         w = self._branch.repository.weave_store.get_weave(file_id, self._branch.repository.get_transaction())
+        
+        revid_set = set()
+        for line_revid, text in w.annotate_iter(file_revid):
+            revid_set.add(line_revid)
+            if self._BADCHARS_RE.match(text):
+                # bail out; this isn't displayable text
+                yield util.Container(parity=0, lineno=1, status='same',
+                                     text='<i>' + util.html_clean('(This is a binary file.)') + '</i>',
+                                     change=util.Container())
+                return
+        change_cache = dict([(c.revid, c) for c in self.get_changes(list(revid_set))])
+        
         last_line_revid = None
         for line_revid, text in w.annotate_iter(file_revid):
             if line_revid == last_line_revid:
@@ -644,16 +794,13 @@
                 status = 'changed'
                 parity ^= 1
                 last_line_revid = line_revid
-                change = revision_cache.get(line_revid, None)
-                if change is None:
-                    change = self.get_change(line_revid)
-                    revision_cache[line_revid] = change
+                change = change_cache[line_revid]
                 trunc_revno = change.revno
                 if len(trunc_revno) > 10:
                     trunc_revno = trunc_revno[:9] + '...'
                 
             yield util.Container(parity=parity, lineno=lineno, status=status,
-                                 trunc_revno=trunc_revno, change=change, text=util.html_clean(text))
+                                 change=change, text=util.html_clean(text))
             lineno += 1
         
-        log.debug('annotate: %r secs' % (time.time() - z,))
+        self.log.debug('annotate: %r secs' % (time.time() - z,))

=== modified file 'loggerhead/static/css/style.css'
--- loggerhead/static/css/style.css	2006-12-15 08:42:22 +0000
+++ loggerhead/static/css/style.css	2006-12-20 00:43:10 +0000
@@ -10,29 +10,40 @@
 }
 
 a {
-	text-decoration: none;
-	color: #000088;
+    text-decoration: none;
+    color: #000088;
 }
 a:hover {
-	color: black;
+    color: black;
 }
 
 img {
-	border: none;
+    border: none;
 }
 
 h1 {
     color: #708a96;
-	font-weight: normal;
-	font-size: 1.6em;
+    font-weight: normal;
+    font-size: 1.6em;
     margin-bottom: 1.5em;
 }
 h1 .filename {
-	font-family: monospace;
+    font-family: monospace;
 }
 
 .branch-name {
-	color: #47626e;
+    color: #47626e;
+}
+
+.footer {
+	margin-top: 1em;
+	margin-botom: 2em;
+}
+.footer .banner {
+	margin-right: 5px;
+}
+.footer .banner img {
+	border: 1px solid #8cacbb;
 }
 
 
@@ -42,7 +53,7 @@
 }
 
 .navbar table, .bar table {
-	width: 100%;
+    width: 100%;
 }
 
 .bar {
@@ -56,12 +67,12 @@
 /* revision navigation buttons (the ones at the top right) */
 .navbuttons {
     padding: 2px;
-	bottom: 0px;
-	text-align: right;
-	width: auto;
-	border-left: 1px dotted #8cacbb;
-	padding-left: 0.5em;
-	padding-right: 0.5em;
+    bottom: 0px;
+    text-align: right;
+    width: auto;
+    border-left: 1px dotted #8cacbb;
+    padding-left: 0.5em;
+    padding-right: 0.5em;
 }
 
 .navbuttons a {
@@ -75,22 +86,11 @@
     color: blue;
 }
 
-/* the "changes: 425 / 425" header at the top right */
-.navposition {
-	margin-bottom: 1.5em;
-}
-
-.navposition table tr td {
-    margin-top: 0px;
-    font-weight: bold;
-    font-size: smaller;
-}
-
 /* navigation button in the pages (the ones at the top left, and bottom) */
 .buttons {
 }
 
-.buttons a {
+.buttons a,.buttons .search {
     font-weight: normal;
     color: #000088;
     font-size: 12pt;
@@ -105,7 +105,7 @@
 }
 
 .feed img {
-	vertical-align: text-top;
+    vertical-align: text-top;
 }
 
 .rbuttons a {
@@ -122,7 +122,7 @@
 /* ----- revision display in various places ----- */
 
 .revision .revno {
-	font-weight: bold;
+    font-weight: bold;
 }
 
 .revision .revid {
@@ -139,13 +139,13 @@
 /* ---- sub-links underneath the headline on some pages ----- */
 
 div.links {
-	padding-top: 0.5em;
-	padding-left: 3em;
-	font-size: 50%;
+    padding-top: 0.5em;
+    padding-left: 3em;
+    font-size: 50%;
 }
 
 div.links div {
-	margin-bottom: 3px;
+    margin-bottom: 3px;
 }
 
 
@@ -168,28 +168,28 @@
     vertical-align: top;
 
     padding-left: 0.5em;
-	padding-right: 0.5em;
-	padding-top: 2px;
-	padding-bottom: 2px;
+    padding-right: 0.5em;
+    padding-top: 2px;
+    padding-bottom: 2px;
 }
 
 .info-entry td {
     font-weight: normal;
-  	padding-left: 0.5em;
-	padding-top: 2px;
-	padding-bottom: 2px;
+    padding-left: 0.5em;
+    padding-top: 2px;
+    padding-bottom: 2px;
 }
 
 
 /* ----- changelog page ----- */
 
 .hide-all {
-	font-size: smaller;
+    font-size: smaller;
 }
 
 /* secret css-jutsu to make hidden margins vanish on a table */
 .log-entries table {
-	border-collapse: collapse;
+    border-collapse: collapse;
     width: 100%;
 }
 
@@ -197,8 +197,8 @@
     background-color: #dee7ec;
     padding: 3px;
     width: 80%;
-	margin-top: 1em;
-	margin-left: 1em;
+    margin-top: 1em;
+    margin-left: 1em;
 }
 
 .log-entries .revision-header td.expand-button {
@@ -206,63 +206,63 @@
 }
 
 .log-entries .revision-header .revision-number {
-	width: 6em;
-	padding-left: 0.5em;
+    width: 6em;
+    padding-left: 0.5em;
 }
 
 .log-entries .revision-header .inventory-link {
-	font-size: smaller;
-	vertical-align: middle;
-	width: 4em;
-	text-align: right;
-	padding-right: 0.5em;
+    font-size: smaller;
+    vertical-align: middle;
+    width: 4em;
+    text-align: right;
+    padding-right: 0.5em;
 }
 
 .log-entries .revision-details-block {
-	width: 80%;
-	border: none;
-	margin-left: 8em;
-	margin-bottom: 1.5em;
+    width: 80%;
+    border: none;
+    margin-left: 8em;
+    margin-bottom: 1.5em;
 }
 
 .log-entries .revision-details table {
-	font-size: smaller;
-	width: 100%;
+    font-size: smaller;
+    width: 100%;
 }
 
 .log-entries .revision-details th {
-	width: 8em;
-	text-align: right;
-	padding-right: 0.5em;
-	vertical-align: top;
+    width: 8em;
+    text-align: right;
+    padding-right: 0.5em;
+    vertical-align: top;
 }
 
 .log-entries .revision-details td.date {
-/*	font-family: monospace; */
+/*    font-family: monospace; */
 }
 
 .log-entries .revision-details th.files {
-	padding-top: 1em;
+    padding-top: 1em;
 }
 
 .log-entries .revision-details td.files {
-	padding-top: 1em;
+    padding-top: 1em;
 }
 
 .log-entries .revision-details td.files .filename {
-	font-family: monospace;
+    font-family: monospace;
 }
 
-.log-entries .revision-details td.files a.jump {
+.log-entries .revision-details td.files .jump {
     font-size: smaller;
 }
 
 .log-entries .revision-details th.description {
-	padding-top: 1em;
+    padding-top: 1em;
 }
 .log-entries .revision-details td.description {
-	font-family: monospace;
-	padding-top: 1em;
+    font-family: monospace;
+    padding-top: 1em;
 }
 
 /* default state: hide all */
@@ -285,46 +285,46 @@
 
 /* secret css-jutsu to make hidden margins vanish on a table */
 .revision-info table {
-	border-collapse: collapse;
+    border-collapse: collapse;
     width: 100%;
 }
 
 .revision-info th {
-	background-color: #dee7ec;
+    background-color: #dee7ec;
     color: black;
-	font-weight: normal;
-	text-align: right;
-	vertical-align: top;
-	
-	padding-left: 0.5em;
-	padding-right: 0.5em;
-	padding-top: 2px;
-	padding-bottom: 2px;
+    font-weight: normal;
+    text-align: right;
+    vertical-align: top;
+    
+    padding-left: 0.5em;
+    padding-right: 0.5em;
+    padding-top: 2px;
+    padding-bottom: 2px;
 }
 
 .revision-info td {
-	font-weight: normal;
-  	padding-left: 0.5em;
-	padding-top: 2px;
-	padding-bottom: 2px;
+    font-weight: normal;
+    padding-left: 0.5em;
+    padding-top: 2px;
+    padding-bottom: 2px;
 }
 
 .revision-info table td.date {
-	font-family: monospace;
+    font-family: monospace;
 }
 
 .revision-info table td.description {
-	font-family: monospace;
+    font-family: monospace;
 }
 
 .revision-info table td.files .filename {
-	font-family: monospace;
+    font-family: monospace;
 }
 
 .revision-info table td.files a.jump {
-	font-size: smaller;
+    font-size: smaller;
 }
-	
+    
 .revision-info table tr.divider {
     border-top: 1px solid #8cacbb;
 }
@@ -335,73 +335,95 @@
 .diff {
     margin-bottom: 1em;
     padding-top: 10px;
-    margin-top: 20px;
     width: 100%;
     background-color: #fcfcfc;
     font-family: monospace;
 }
 
+
 .diff-block,.diff-chunk {
-	border-collapse: collapse;
+    border-collapse: collapse;
     width: 100%;
+    margin-bottom: 1em;
 }
 
 .diff-block th.filename {
-	font-family: monospace;
-	font-weight: normal;
-	font-size: smaller;
-	padding-left: 1em;
-	padding-top: 1em;
-	padding-bottom: 1em;
-	text-align: left;
-	background-color: #dee7ec;
+    font-family: monospace;
+    font-weight: normal;
+    font-size: smaller;
+    padding-left: 1em;
+    padding-top: 1em;
+    padding-bottom: 1em;
+    text-align: left;
+    background-color: #dee7ec;
 }
 
 .diff-chunk {
     font-size: 66%;
     margin-top: 0px;
-	margin-bottom: 1em;
+    margin-bottom: 2em;
 }
 
 .diff-chunk th.lineno {
-	font-weight: bold;
+    font-weight: bold;
     border-bottom: 1px solid #8cacbb;
     border-top: 1px solid #dee7ec;
     border-left: 1px solid #dee7ec;
 }
-.diff-chunk .insert {
-	background-color: #b0efb0;
+.diff-insert {
+    background-color: #b0efb0;
 }
-.diff-chunk .delete {
-	background-color: #ffa0a0;
+.diff-delete {
+    background-color: #ffa0a0;
 }
 .diff-chunk td.text {
-	padding-left: 1em;
+    padding-left: 1em;
+}
+
+.diff-key {
+    margin-top: 20px;
+    font-size: smaller;
+    line-height: 1em;
+    padding-left: 1em;
+}
+
+.diff-key-block {
+    width: 0.7em;
+    height: 0.7em;
+    border: 1px solid #8cacbb;
+    margin-right: 5em;
+}
+
+.diff-key .label {
+    position: relative;
+    left: 1.3em;
+    top: -1em;
+    line-height: 1em;
 }
 
 .diff .lineno {
-	color: purple;
-	background-color: #dee7ec; 
-	width: 4em;
-	text-align: right;
-	padding-right: 3px;
-	border-bottom: 1px solid #8cacbb;
+    color: purple;
+    background-color: #dee7ec; 
+    width: 4em;
+    text-align: right;
+    padding-right: 3px;
+    border-bottom: 1px solid #8cacbb;
 }
 .diff .info {
-	color: blue;
-	background-color: #dee7ec;
-	margin-top: 15px;
+    color: blue;
+    background-color: #dee7ec;
+    margin-top: 15px;
 }
 
 
 /* ----- file inventory ----- */
 
 .inventory-path {
-	margin-top: 2em;
+    margin-top: 2em;
 }
 
 table.inventory {
-	border-collapse: collapse;
+    border-collapse: collapse;
     width: 80%;
     font-size: smaller;
     margin-top: 1em;
@@ -409,20 +431,20 @@
 }
 
 table.inventory td {
-	padding-top: 3px;
-	padding-bottom: 3px;
-	vertical-align: middle;
+    padding-top: 3px;
+    padding-bottom: 3px;
+    vertical-align: middle;
 }
 
 table.inventory th {
     background-color: #8cacbb;
     padding-top: 3px;
     padding-bottom: 3px;
-	text-align: left;
+    text-align: left;
 }
 
 table.inventory th.permissions {
-	padding-left: 0.5em;
+    padding-left: 0.5em;
 }
 
 .parity0 { background-color: #dee7ec; }
@@ -430,24 +452,30 @@
 
 .inventory td.permissions {
     font-family: monospace;
-    width: 10em;
-	padding-left: 0.5em;
+    width: 9em;
+    padding-left: 0.5em;
 }
 
 .inventory td.filename {
     font-family: monospace;
 }
 
+.inventory td.size {
+	font-family: monospace;
+	font-size: smaller;
+	width: 5em;
+}
+
 .inventory td.revision {
     width: 8em;
     padding-right: 1.5em;
 }
 
 .inventory td.date {
-	font-family: monospace;
-	font-size: smaller;
-	width: 10em;
-	padding-right: 1.5em;
+    font-family: monospace;
+    font-size: smaller;
+    width: 10em;
+    padding-right: 1.5em;
 }
 
 .inventory td.changes-link {
@@ -455,6 +483,11 @@
     font-size: smaller;
 }
 
+.inventory td.download-link {
+    width: 8em;
+    font-size: smaller;
+}
+
 span.folder {
     font-family: monospace;
 }
@@ -463,36 +496,36 @@
 /* ----- file annotation ----- */
 
 .annotate-path {
-	font-family: monospace;
+    font-family: monospace;
 }
 
 .annotate table {
-	border-collapse: collapse;
+    border-collapse: collapse;
     width: 100%;
     font-size: 66%;
 }
 
 .annotate td {
-	padding-top: 2px;
-	padding-bottom: 2px;
-	vertical-align: bottom;
+    padding-top: 2px;
+    padding-bottom: 2px;
+    vertical-align: bottom;
 }
 
 .annotate td.lineno {
-	font-family: monospace;
-	text-align: right;
-	width: 4em;
-	padding-right: 5px;
-	font-weight: bold;
-	border-bottom: 1px solid #8cacbb;
+    font-family: monospace;
+    text-align: right;
+    width: 4em;
+    padding-right: 5px;
+    font-weight: bold;
+    border-bottom: 1px solid #8cacbb;
 }
 
 .annotate td.revno {
-	font-family: monospace;
-	text-align: right;
-	width: 8em;
-	padding-right: 5px;
-	border-right: 1px solid #8cacbb;
+    font-family: monospace;
+    text-align: right;
+    width: 8em;
+    padding-right: 5px;
+    border-right: 1px solid #8cacbb;
 }
 
 .annotate td.text {
@@ -504,7 +537,7 @@
     background-color: #8cacbb;
     padding-top: 3px;
     padding-bottom: 3px;
-	text-align: left;
+    text-align: left;
 }
 
 .annotate th.lineno {
@@ -517,5 +550,5 @@
 }
 
 .annotate th.text {
-	padding-left: 1em;
+    padding-left: 1em;
 }

=== added file 'loggerhead/static/images/bazaar-banner.png'
Binary files loggerhead/static/images/bazaar-banner.png and loggerhead/static/images/bazaar-banner.png differ
=== modified file 'loggerhead/static/images/favicon.ico'
Binary files loggerhead/static/images/favicon.ico and loggerhead/static/images/favicon.ico differ
=== added file 'loggerhead/static/images/loggerhead-banner.png'
Binary files loggerhead/static/images/loggerhead-banner.png and loggerhead/static/images/loggerhead-banner.png differ
=== modified file 'loggerhead/templates/annotate.kid'
--- loggerhead/templates/annotate.kid	2006-12-15 08:42:22 +0000
+++ loggerhead/templates/annotate.kid	2006-12-20 09:29:42 +0000
@@ -3,56 +3,22 @@
     py:extends="'master.kid'">
 <head>
     <meta content="text/html; charset=utf-8" http-equiv="Content-Type" py:replace="''"/>
-    <title> ${branch_name} : contents of ${path} at revision ${change.revno} </title>
+    <title> ${branch.friendly_name} : contents of ${path} at revision ${change.revno} </title>
 </head>
 
 <body>
 
 ${navbar()}
 
-<h1> <span class="branch-name">${branch_name}</span> : <span class="annotate-path">${path}</span> (revision ${change.revno})
+<h1> <span class="branch-name">${branch.friendly_name}</span> : <span class="annotate-path">${path}</span> (revision ${change.revno})
 	<div class="links">
-	    <div> <b>&#8594;</b> <a href="${tg.url([ '/files', revid ])}"> browse files </a> </div>
-	    <div> <b>&#8594;</b> <a href="${tg.url('/revision', start_revid=revid)}"> view revision </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url([ '/files', revid ])}"> browse files </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url('/revision', start_revid=revid)}"> view revision </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url('/changes', start_revid=revid, file_id=file_id)}"> view changes to this file </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url([ '/download', revid, file_id, filename ])}"> view/download file </a> </div>
 	</div>
 </h1>
 
-<!-- !FIXME: this is just copied verbatim from revision.kid -->
-<!--div class="revision-info">
-    <table>
-        <tr>
-            <th class="author">committed by:</th>
-            <td class="author"> ${util.hide_email(change.author)} </td>
-        </tr>
-        <tr>
-            <th class="date">date:</th>
-            <td class="date"> ${change.date.strftime('%d %b %Y %H:%M')} </td>
-        </tr>
-
-        <tr py:if="len(change.merge_points) > 0">
-            <th class="children"> merged in: </th>
-            <td class="children">
-                <span py:for="child in change.merge_points">
-                    ${revlink_path(child.revid, child.revid, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')', None)} &nbsp;
-                </span>
-            </td>
-        </tr>
-        <tr py:if="len(change.parents) > 1">
-        	<th class="parents"> merged from: </th>
-        	<td class="parents">
-        	    <span py:for="parent in change.parents"><span py:if="parent.revid != change.parents[0].revid">
-           	        ${revlink_path(parent.revid, parent.revid, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')', None)} &nbsp;
-                </span></span>
-        	</td>
-        </tr>
-
-        <tr>
-            <th class="description">description:</th>
-            <td class="description"><span py:for="line in change.comment_clean">${XML(line)} <br /></span></td>
-        </tr>
-    </table>
-</div-->
-
 <div class="annotate">
     <table>
         <tr>
@@ -64,10 +30,9 @@
         <tr py:for="line in contents" class="parity${line.parity}">
             <td class="lineno ${line.status}"> ${line.lineno} </td>
             <td class="revno ${line.status}">
-                <a py:if="line.status=='changed'" href="${tg.url('/revision', start_revid=line.change.revid, path=path)}"
-                    title="${line.change.revno} by ${util.hide_email(line.change.author)}, on ${line.change.date.strftime('%d %b %Y %H:%M')} (${util.ago(line.change.date)})">${line.trunc_revno}</a>
+                <a py:if="line.status=='changed'" href="${branch.url('/revision', start_revid=line.change.revid, file_id=file_id)}"
+                    title="${line.change.revno} by ${util.hide_email(line.change.author)}, on ${line.change.date.strftime('%d %b %Y %H:%M')} (${util.ago(line.change.date)})">${util.trunc(line.change.revno)}</a>
             </td>
-            <!--td class="author"> ${util.hide_email(line.author)} </td-->
             <td class="text ${line.status}"> ${XML(line.text)} </td>
         </tr>
     </table>

=== modified file 'loggerhead/templates/atom.kid'
--- loggerhead/templates/atom.kid	2006-12-11 06:44:19 +0000
+++ loggerhead/templates/atom.kid	2006-12-20 09:29:42 +0000
@@ -1,18 +1,18 @@
 <?xml version="1.0" encoding="utf-8"?>
 <feed xmlns="http://www.w3.org/2005/Atom" xmlns:py="http://purl.org/kid/ns#">
-    <title> bazaar changes for ${branch_name} </title>
+    <title> bazaar changes for ${branch.friendly_name} </title>
     <updated>${updated}</updated>
-    <id>${tg.url([ external_url, 'atom' ])}</id>
-    <link rel="self" href="${tg.url([ external_url, 'atom' ])}" />
+    <id>${tg.url([ external_url, branch.name, 'atom' ])}</id>
+    <link rel="self" href="${tg.url([ external_url, branch.name, 'atom' ])}" />
 
 	<entry py:for="entry in changes">
-	    <title> ${entry.short_comment} </title>
-	    <updated> ${entry.date.isoformat() + 'Z'} </updated>
-	    <id>${tg.url([ external_url, 'atom', entry.revid ])}</id>
+	    <title> ${entry.revno}: ${entry.short_comment} </title>
+	    <updated>${entry.date.isoformat() + 'Z'}</updated>
+	    <id>${tg.url([ external_url, 'revision', entry.revid ])}</id>
 	    <author> <name> ${util.hide_email(entry.author)} </name> </author>
 	    <content type="text">
             ${entry.comment}
 	    </content>
-	    <link rel="alternate" href="${tg.url([ external_url, 'revision', entry.revid ])}" />
+	    <link rel="alternate" href="${tg.url([ external_url, branch.name, 'revision', entry.revid ])}" />
 	</entry>
 </feed>

=== modified file 'loggerhead/templates/changelog.kid'
--- loggerhead/templates/changelog.kid	2006-12-15 08:42:22 +0000
+++ loggerhead/templates/changelog.kid	2006-12-20 09:29:42 +0000
@@ -3,15 +3,15 @@
     py:extends="'master.kid'">
 <head>
     <meta content="text/html; charset=utf-8" http-equiv="Content-Type" py:replace="''"/>
-    <title> ${branch_name} : changes </title>
-    <link rel="alternate" type="application/atom+xml" href="${tg.url('/atom')}" title="RSS feed for ${branch_name}" />
+    <title> ${branch.friendly_name} : changes </title>
+    <link rel="alternate" type="application/atom+xml" href="${branch.url('/atom')}" title="RSS feed for ${branch.friendly_name}" />
     
     <span py:def="loglink(revid, text)">
         <a title="Show history" href="${tg.url('/changes', start_revid=revid)}" class="revlink"> ${text} </a>
     </span>
 
-    <span py:def="file_link(filename, revid)">
-        <a href="${tg.url([ '/annotate', revid ], path=filename)}" title="Annotate ${filename}">${filename}</a>
+    <span py:def="file_link(filename, file_id, revid)">
+        <a href="${tg.url([ '/annotate', revid ], file_id=file_id)}" title="Annotate ${filename}">${filename}</a>
     </span>
 
     <!-- this is totally matty's fault.  i don't like javacsript. ;) -->
@@ -49,13 +49,20 @@
 
 ${navbar()}
 
-<h1> <span class="branch-name">${branch_name}</span> : changes
-<span py:if="path"> to <span class="filename">${path}</span></span>
-<span py:if="last_revid != start_revid"> from ${history.get_revno(start_revid)} </span>
+<h1> <span class="branch-name">${branch.friendly_name}</span> : changes
+<span py:if="file_id"> to <span class="filename">${history.get_path(revid, file_id)}</span></span>
+<span py:if="viewing_from"> from ${history.get_revno(start_revid)} </span>
+<span py:if="query"> matching "${query}"</span>
 </h1>
 
+<span py:if="search_failed">
+    Sorry, no results found for your search.
+</span>
+
+<span py:if="not search_failed">
 <a class="hide-all" id="hide-all" href="javascript:displayAll('none', '')"> (hide all) </a>
 <a class="hide-all" id="show-all" href="javascript:displayAll('', 'none')"> (show all) </a>
+</span>
 
 <div class="log-entries">
     <div py:for="entry in changes" class="revision">
@@ -63,7 +70,7 @@
         <div class="revision-header">
             <table>
                 <tr>
-                    <td class="revision-number"> ${revlink_path(entry.revid, start_revid, entry.revno, path)} </td>
+                    <td class="revision-number"> ${revlink_q(entry.revid, start_revid, file_id, query, util.trunc(entry.revno))} </td>
                     <td class="expand-button">
                         <a href="javascript:displayDetails('${entry.revno}', 'none', '')" id="hide-${entry.revno}" class="show-button">
                             <img src="${tg.url('/static/images/nav-small-down.gif')}" width="10" height="10" />
@@ -72,7 +79,7 @@
                         	<img src="${tg.url('/static/images/nav-small-right.gif')}" witdh="10" height="10" />
                         </a>
                     </td>
-					<td class="summary"> ${revlink_path(entry.revid, start_revid, entry.short_comment, path)} </td>
+					<td class="summary"> ${revlink_q(entry.revid, start_revid, file_id, query, entry.short_comment)} </td>
 					<td class="inventory-link"> <a href="${tg.url([ '/files', entry.revid ])}">&#8594; files</a> </td>
 				</tr>
 			</table>
@@ -97,7 +104,7 @@
 			            <th class="children"> merged in: </th>
 			            <td class="children">
 			                <span py:for="child in entry.merge_points">
-			                    ${loglink(child.revid, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')')} &nbsp; 
+			                    ${loglink(child.revid, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')')} <br />
 			                </span>
 			            </td>
 			        </tr>
@@ -105,30 +112,30 @@
 			        	<th class="parents"> merged from: </th>
 			        	<td class="parents">
 			        	    <span py:for="parent in entry.parents"><span py:if="parent.revid != entry.parents[0].revid">
-			        	        ${loglink(parent.revid, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')')} &nbsp;
+			        	        ${loglink(parent.revid, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')')} <br />
 			        	    </span></span>
 			        	</td>
 			        </tr>
 
 			        <tr py:if="entry.changes.added">
 			            <th class="files"> files added: </th>
-			            <td class="files"> <span py:for="filename in entry.changes.added" class="filename">${file_link(filename, entry.revid)} <br /></span> </td>
+			            <td class="files"> <span py:for="filename, fid in entry.changes.added" class="filename">${file_link(filename, fid, entry.revid)} <br /></span> </td>
 			        </tr>
 			        <tr py:if="entry.changes.removed">
 			            <th class="files"> files removed: </th>
-			            <td class="files"> <span py:for="filename in entry.changes.removed" class="filename">${file_link(filename, entry.revid)} <br /></span> </td>
+			            <td class="files"> <span py:for="filename, fid in entry.changes.removed" class="filename">${file_link(filename, fid, entry.revid)} <br /></span> </td>
 			        </tr>
 			        <tr py:if="entry.changes.renamed">
 			            <th class="files"> files renamed: </th>
-			            <td class="files"> <span py:for="old_filename, new_filename in entry.changes.renamed" class="filename">
-			                ${file_link(old_filename, entry.revid)} => ${file_link(new_filename, entry.revid)}<br />
+			            <td class="files"> <span py:for="old_filename, new_filename, fid in entry.changes.renamed" class="filename">
+			                ${file_link(old_filename, fid, entry.revid)} => ${file_link(new_filename, fid, entry.revid)}<br />
 			            </span> </td>
 			        </tr>
 			        <tr py:if="entry.changes.modified">
 			            <th class="files"> files modified: </th>
 			            <td class="files"> <span py:for="item in entry.changes.modified">
-			                <span class="filename">${file_link(item.filename, entry.revid)}</span> &nbsp;
-			                <a href="${tg.url([ '/revision', entry.revid ], start_revid=start_revid, path=path) + '#' + item.filename}" class="jump">(jump to diff)</a>
+			                <span class="filename">${file_link(item.filename, item.file_id, entry.revid)}</span> &nbsp;
+			                <a href="${tg.url([ '/revision', entry.revid ], start_revid=start_revid, file_id=file_id, query=query) + '#' + item.filename}" class="jump">&#8594; diff</a>
 			                <br />
 			            </span> </td>
 			        </tr>
@@ -142,7 +149,7 @@
     </div>
 </div>
 
-<div class="bar">
+<div py:if="navigation.prev_page_revid or navigation.next_page_revid" class="bar">
     <table>
         <tr>
         	<td class="buttons">

=== modified file 'loggerhead/templates/inventory.kid'
--- loggerhead/templates/inventory.kid	2006-12-15 08:42:22 +0000
+++ loggerhead/templates/inventory.kid	2006-12-20 09:29:42 +0000
@@ -3,17 +3,17 @@
     py:extends="'master.kid'">
 <head>
     <meta content="text/html; charset=utf-8" http-equiv="Content-Type" py:replace="''"/>
-    <title> ${branch_name} : files for revision ${change.revno} </title>
+    <title> ${branch.friendly_name} : files for revision ${change.revno} </title>
 </head>
 
 <body>
 
 ${navbar()}
 
-<h1> <span class="branch-name">${branch_name}</span> : files for revision ${change.revno}
+<h1> <span class="branch-name">${branch.friendly_name}</span> : files for revision ${change.revno}
 	<div class="links">
-	    <div> <b>&#8594;</b> <a href="${tg.url('/revision', start_revid=revid)}"> view revision </a> </div>
-	    <div> <b>&#8594;</b> <a href="${tg.url('/changes', start_revid=revid)}"> view branch changes </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url('/revision', start_revid=revid)}"> view revision </a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url('/changes', start_revid=revid)}"> view branch changes </a> </div>
 	</div>
 </h1>
 
@@ -33,7 +33,7 @@
             <th class="children"> merged in: </th>
             <td class="children">
                 <span py:for="child in change.merge_points">
-                    ${revlink_path(child.revid, child.revid, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')', None)} &nbsp; 
+                    ${revlink(child.revid, child.revid, None, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')')} <br />
                 </span>
             </td>
         </tr>
@@ -41,7 +41,7 @@
         	<th class="parents"> merged from: </th>
         	<td class="parents">
         	    <span py:for="parent in change.parents"><span py:if="parent.revid != change.parents[0].revid">
-        	        ${revlink_path(parent.revid, parent.revid, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')', None)} &nbsp;
+        	        ${revlink(parent.revid, parent.revid, None, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')')} <br />
         	    </span></span>
         	</td>
         </tr>
@@ -60,28 +60,32 @@
 <table class="inventory" width="100%">
     <tr class="header">
         <th class="permissions"> Permissions </th>
-        <th> Filename </th>
+        <th> <a href="${branch.url([ '/files', revid ], file_id=file_id, sort='filename')}">Filename</a> </th>
+        <th> <a href="${branch.url([ '/files', revid ], file_id=file_id, sort='size')}">Size</a> </th>
         <th> Last change </th>
-        <th> When </th>
+        <th> <a href="${branch.url([ '/files', revid ], file_id=file_id, sort='date')}">When</a> </th>
         <th> History </th>
+        <th> Download </th>
     </tr>
     
     <tr class="parity1" py:if="updir">
         <td class="permissions">drwxr-xr-x</td>
-        <td class="filename directory"><a href="${tg.url([ '/files', revid ], path=updir)}"> (up) </a></td>
-        <td> </td> <td> </td> <td> </td>
+        <td class="filename directory"><a href="${branch.url([ '/files', revid ], file_id=updir_file_id)}"> (up) </a></td>
+        <td> </td> <td> </td> <td> </td> <td> </td>
     </tr>
 
     <tr py:for="file in filelist" class="parity${file.parity}">
         <td class="permissions"> ${util.fake_permissions(file.kind, file.executable)} </td>
         <td class="filename ${file.kind}">
-            <a py:if="file.kind=='directory'" href="${tg.url([ '/files', revid ], path=posixpath.join(path, file.pathname))}">${file.filename}/</a>
+            <a py:if="file.kind=='directory'" href="${branch.url([ '/files', revid ], file_id=file.file_id)}">${file.filename}/</a>
             <span py:if="file.kind=='symlink'">${file.filename}@</span>
-            <a py:if="file.kind=='file'" href="${tg.url([ '/annotate', revid ], path=posixpath.join(path, file.pathname))}">${file.filename}</a>
+            <a py:if="file.kind=='file'" href="${branch.url([ '/annotate', revid ], file_id=file.file_id)}">${file.filename}</a>
         </td>
-        <td class="revision"> ${revlink_path(file.revid, file.revid, util.trunc(file.change.revno, 15), posixpath.join(path, file.pathname))} </td>
+        <td class="size"> <span py:if="file.kind=='file'"> ${util.human_size(file.size)} </span></td>
+        <td class="revision"> ${revlink(file.revid, file.revid, file.file_id, util.trunc(file.change.revno, 15))} </td>
         <td class="date"> ${file.change.date.strftime('%d %b %Y %H:%M')} </td>
-        <td class="changes-link"> <a href="${tg.url('/changes', start_revid=file.revid, path=posixpath.join(path, file.pathname))}"> &#8594; changes </a></td>
+        <td class="changes-link"> <a href="${branch.url('/changes', start_revid=file.revid, file_id=file.file_id)}"> &#8594; changes </a></td>
+        <td class="download-link"> <a href="${branch.url([ '/download', file.revid, file.file_id, file.filename ])}"> &#8594; download </a></td>
     </tr>
 </table>
 

=== modified file 'loggerhead/templates/master.kid'
--- loggerhead/templates/master.kid	2006-12-15 10:05:12 +0000
+++ loggerhead/templates/master.kid	2006-12-20 09:29:42 +0000
@@ -12,19 +12,22 @@
 
 <!-- !define common navbar -->
 <span py:def="navbar()">
-    <!-- !requires: ${navigation: start_revid, revid, revlist, pagesize, buttons, scan_url}, ${history} -->
+    <!-- !requires: ${navigation: start_revid, revid, revid_list, pagesize, buttons, scan_url}, ${branch}, ${history} -->
     <div class="navbar">
         <div class="bar">
+            <!-- form must go OUTSIDE the table, or safari will add extra padding :( -->
+            <form action="${branch.url('/changes', start_revid=getattr(navigation, 'start_revid', None), file_id=getattr(navigation, 'file_id', None))}">
             <table>
                 <tr><td>
                     <span class="buttons">
                         <!-- ! navbar buttons never change, from now on.  i decree it! -->
-                        <a href="${tg.url('/changes')}"> changes </a>
-                        <a href="${tg.url('/files')}"> files </a>
+                        <a href="${branch.url('/changes')}"> changes </a>
+                        <a href="${branch.url('/files')}"> files </a>
+                        <span class="search"> search: <input type="text" name="q" /> </span>
                     </span>
-                </td><td align="right" py:if="hasattr(navigation, 'revlist')">
+                </td><td align="right" py:if="hasattr(navigation, 'revid_list')">
                     <span py:if="hasattr(navigation, 'feed')" class="rbuttons feed">
-                        <a href="${tg.url('/atom')}"><img src="/static/images/feed-icon-16x16.gif" /></a>
+                        <a href="${branch.url('/atom')}"><img src="${tg.url('/static/images/feed-icon-16x16.gif')}" /></a>
                     </span>
                     <span class="navbuttons">
                     	<span py:if="navigation.prev_page_revid"> <a href="${navigation.prev_page_url}" title="Previous page"> &#171; </a>	</span>
@@ -35,15 +38,16 @@
                     </span>
                 </td></tr>
             </table>
+            </form>
         </div>
     </div>
 </span>
 
-<span py:def="revlink(revid, text)">
-    <a title="Show revision" href="${tg.url([ '/revision', revid ])}" class="revlink"> ${text} </a>
+<span py:def="revlink(revid, start_revid, file_id, text)">
+    <a title="Show revision ${history.get_revno(revid)}" href="${branch.url([ '/revision', revid ], start_revid=start_revid, file_id=file_id)}" class="revlink"> ${text} </a>
 </span>
-<span py:def="revlink_path(revid, start_revid, text, path)">
-    <a title="Show revision" href="${tg.url([ '/revision', revid ], start_revid=start_revid, path=path)}" class="revlink"> ${text} </a>
+<span py:def="revlink_q(revid, start_revid, file_id, query, text)">
+    <a title="Show revision ${history.get_revno(revid)}" href="${branch.url([ '/revision', revid ], start_revid=start_revid, file_id=file_id, q=query)}" class="revlink"> ${text} </a>
 </span>
 
 </head>
@@ -53,6 +57,11 @@
     <div py:replace="[item.text]+item[:]"/>
 
 	<!-- End of main_content -->
+	
+	<div class="footer">
+	    <span class="banner"><a href="http://www.bazaar-vcs.org/"><img src="${tg.url('/static/images/bazaar-banner.png')}" /></a></span>
+	    <span class="banner"><a href="http://www.lag.net/loggerhead/"><img src="${tg.url('/static/images/loggerhead-banner.png')}" /></a></span>
+	</div>
 </body>
 
 </html>

=== modified file 'loggerhead/templates/revision.kid'
--- loggerhead/templates/revision.kid	2006-12-15 08:42:22 +0000
+++ loggerhead/templates/revision.kid	2006-12-20 09:29:42 +0000
@@ -3,10 +3,10 @@
     py:extends="'master.kid'">
 <head>
     <meta content="text/html; charset=utf-8" http-equiv="Content-Type" py:replace="''"/>
-    <title> ${branch_name} : revision ${change.revno} </title>
+    <title> ${branch.friendly_name} : revision ${change.revno} </title>
     
-    <span py:def="file_link(filename)">
-        <a href="${tg.url([ '/annotate', revid ], path=filename)}" title="Annotate ${filename}">${filename}</a>
+    <span py:def="file_link(filename, file_id)">
+        <a href="${branch.url([ '/annotate', revid ], file_id=file_id)}" title="Annotate ${filename}">${filename}</a>
     </span>
 </head>
 
@@ -14,10 +14,10 @@
 
 ${navbar()}
 
-<h1> <span class="branch-name">${branch_name}</span> : revision ${change.revno}
+<h1> <span class="branch-name">${branch.friendly_name}</span> : revision ${change.revno}
 	<div class="links">
-	    <div> <b>&#8594;</b> <a href="${tg.url([ '/files', revid ])}">browse files</a> </div>
-	    <div> <b>&#8594;</b> <a href="${tg.url('/changes', start_revid=revid)}">view branch changes</a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url([ '/files', revid ])}">browse files</a> </div>
+	    <div> <b>&#8594;</b> <a href="${branch.url('/changes', start_revid=revid)}">view branch changes</a> </div>
 	</div>
 </h1>
  
@@ -36,7 +36,7 @@
             <th class="children"> merged in: </th>
             <td class="children">
                 <span py:for="child in change.merge_points">
-                    ${revlink_path(child.revid, child.revid, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')', None)} &nbsp; 
+                    ${revlink(child.revid, child.revid, None, '(' + child.revno + util.if_present(' %s', child.branch_nick) + ')')} <br /> 
                 </span>
             </td>
         </tr>
@@ -44,7 +44,7 @@
         	<th class="parents"> merged from: </th>
         	<td class="parents">
         	    <span py:for="parent in change.parents"><span py:if="parent.revid != change.parents[0].revid">
-        	        ${revlink_path(parent.revid, parent.revid, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')', None)} &nbsp;
+        	        ${revlink(parent.revid, parent.revid, None, '(' + parent.revno + util.if_present(' %s', parent.branch_nick) + ')')} <br />
         	    </span></span>
         	</td>
         </tr>
@@ -58,44 +58,51 @@
         
         <tr py:if="change.changes.added">
             <th class="files"> files added: </th>
-            <td class="files"> <span py:for="filename in change.changes.added" class="filename">${file_link(filename)} <br /></span> </td>
+            <td class="files"> <span py:for="filename, file_id in change.changes.added" class="filename">${file_link(filename, file_id)} <br /></span> </td>
         </tr>
         <tr py:if="change.changes.removed">
             <th class="files"> files removed: </th>
-            <td class="files"> <span py:for="filename in change.changes.removed" class="filename">${file_link(filename)} <br /></span> </td>
+            <td class="files"> <span py:for="filename, file_id in change.changes.removed" class="filename">${file_link(filename, file_id)} <br /></span> </td>
         </tr>
         <tr py:if="change.changes.renamed">
             <th class="files"> files renamed: </th>
-            <td class="files"> <span py:for="old_filename, new_filename in change.changes.renamed" class="filename">${file_link(old_filename)} => ${file_link(new_filename)}<br /></span> </td>
+            <td class="files"> <span py:for="old_filename, new_filename, file_id in change.changes.renamed" class="filename">
+                ${file_link(old_filename, file_id)} => ${file_link(new_filename, file_id)}<br />
+            </span> </td>
         </tr>
         <tr py:if="change.changes.modified">
             <th class="files"> files modified: </th>
             <td class="files">
                 <span py:for="item in change.changes.modified">
-                    <span class="filename">${file_link(item.filename)}</span> &nbsp; <a href="#${item.filename}" class="jump">(jump to diff)</a><br />
+                    <span class="filename">${file_link(item.filename, item.file_id)}</span> &nbsp; <a href="#${item.filename}" class="jump">&#8594; diff</a><br />
                 </span>
             </td>
         </tr>
     </table>
 </div>
 
+<table class="diff-key" py:if="change.changes.modified"><tr>
+    <td> <div class="diff-key-block diff-insert"></div> <span class="label"> added </span> </td>
+	<td> <div class="diff-key-block diff-delete"></div> <span class="label"> removed </span> </td>
+</tr></table>
+
 <div class="diff" py:if="change.changes.modified">
     <table py:for="item in change.changes.modified" class="diff-block">
-        <tr><th class="filename"> <a href="${tg.url([ '/annotate', change.revid ], path=item.filename)}" name="${item.filename}">${item.filename}</a> </th></tr>
+        <tr><th class="filename"> <a href="${branch.url([ '/annotate', change.revid ], file_id=item.file_id)}" name="${item.filename}">${item.filename}</a> </th></tr>
         <tr><td>
             <table py:for="chunk in item.chunks" class="diff-chunk">
                 <tr> <th class="lineno">old</th> <th class="lineno">new</th> <th></th> </tr>
                 <tr py:for="line in chunk.diff">
                     <td class="lineno">${line.old_lineno}</td>
                     <td class="lineno">${line.new_lineno}</td>
-                    <td class="${line.type} text">${XML(line.line)}</td>
+                    <td class="diff-${line.type} text">${XML(line.line)}</td>
                 </tr>
             </table>
         </td></tr>
     </table>
 </div>
 
-<div class="bar">
+<div py:if="navigation.prev_page_revid or navigation.next_page_revid" class="bar">
     <table>
         <tr>
         	<td class="buttons">

=== added file 'loggerhead/textindex.py'
--- loggerhead/textindex.py	1970-01-01 00:00:00 +0000
+++ loggerhead/textindex.py	2006-12-20 09:29:42 +0000
@@ -0,0 +1,198 @@
+#
+# Copyright (C) 2006  Robey Pointer <robey@lag.net>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+"""
+indexing of the comment text of revisions, for fast searching.
+
+two separate 'shelve' files are created:
+
+    - recorded: revid -> 1 (if the revid is indexed)
+    - index: 3-letter substring -> list(revids)
+"""
+
+import logging
+import os
+import re
+import shelve
+import threading
+import time
+
+from loggerhead import util
+from loggerhead.util import decorator
+
+# if any substring index reaches this many revids, replace the entry with
+# an ALL marker -- it's not worth an explicit index.
+ALL_THRESHOLD = 1000
+ALL = 'ALL'
+
+
+with_lock = util.with_lock('_lock')
+
+
+def normalize_string(s):
+    """
+    remove any punctuation and normalize all whitespace to a single space.
+    """
+    s = util.to_utf8(s).lower()
+    # remove apostrophes completely.
+    s = re.sub(r"'", '', s)
+    # convert other garbage into space
+    s = re.sub(r'[^\w\d]', ' ', s)
+    # compress multiple spaces into one.
+    s = re.sub(r'\s{2,}', ' ', s)
+    # and finally remove leading/trailing whitespace
+    s = s.strip()
+    return s
+
+
+class TextIndex (object):
+    def __init__(self, history, cache_path):
+        self.history = history
+        self.log = history.log
+        
+        if not os.path.exists(cache_path):
+            os.mkdir(cache_path)
+        
+        recorded_filename = os.path.join(cache_path, 'textindex-recorded')
+        index_filename = os.path.join(cache_path, 'textindex')
+        
+        self._recorded = shelve.open(recorded_filename, 'c', protocol=2)
+        self._index = shelve.open(index_filename, 'c', protocol=2)
+        
+        self._lock = threading.RLock()
+        
+        self.log.info('Using search index; %d entries.', len(self._recorded))
+    
+    @with_lock
+    def is_indexed(self, revid):
+        return self._recorded.get(util.to_utf8(revid), None) is not None
+    
+    @with_lock
+    def __len__(self):
+        return len(self._recorded)
+
+    @with_lock
+    def close(self):
+        self._recorded.close()
+        self._index.close()
+    
+    @with_lock
+    def flush(self):
+        self._recorded.sync()
+        self._index.sync()
+    
+    @with_lock
+    def full(self):
+        return (len(self._recorded) >= len(self.history.get_revision_history())) and (util.to_utf8(self.history.last_revid) in self._recorded)
+
+    @with_lock
+    def index_change(self, change):
+        """
+        currently, only indexes the 'comment' field.
+        """
+        comment = normalize_string(change.comment)
+        if len(comment) < 3:
+            return
+        for i in xrange(len(comment) - 2):
+            sub = comment[i:i + 3]
+            revid_set = self._index.get(sub, None)
+            if revid_set is None:
+                revid_set = set()
+            elif revid_set == ALL:
+                # this entry got too big
+                continue
+            revid_set.add(change.revid)
+            if len(revid_set) > ALL_THRESHOLD:
+                revid_set = ALL
+            self._index[sub] = revid_set
+        
+        self._recorded[util.to_utf8(change.revid)] = True
+        return
+    
+    @with_lock
+    def find(self, text, revid_list=None):
+        text = normalize_string(text)
+        if len(text) < 3:
+            return []
+
+        total_set = None
+        if revid_list is not None:
+            total_set = set(revid_list)
+        seen_all = False
+        
+        for i in xrange(len(text) - 2):
+            sub = text[i:i + 3]
+            revid_set = self._index.get(sub, None)
+            if revid_set is None:
+                # zero matches, stop here.
+                return []
+            if revid_set == ALL:
+                # skip
+                seen_all = True
+                continue
+            if total_set is None:
+                total_set = revid_set
+            else:
+                total_set.intersection_update(revid_set)
+            if len(total_set) == 0:
+                return []
+        
+        # tricky: if seen_all is True, one of the substring indices was ALL
+        # (in other words, unindexed), so our results are actually a superset
+        # of the exact answer.
+        #
+        # if we cared, we could do a direct match on the result set and cull
+        # out any that aren't actually matches.  for now, i'm gonna say that
+        # we DON'T care, and if one of the substrings hit ALL, there's a small
+        # chance that we'll give a few false positives, and we don't care.
+        return total_set
+    
+    def check_rebuild(self, max_time=3600):
+        """
+        check if there are any un-indexed revisions, and if so, index them.
+        but don't spend longer than C{max_time} on it.
+        """
+        if self.full():
+            # all done
+            return
+
+        self.log.info('Building search index...')
+        work = list(self.history.get_revision_history())
+        start_time = time.time()
+        last_update = time.time()
+        count = 0
+    
+        for revid in work:
+            if not self.is_indexed(revid):
+                self.index_change(self.history.get_changes([ revid ])[0])
+
+            count += 1
+            now = time.time()
+            if now - start_time > 3600:
+                # there's no point working for hours.  eventually we might even
+                # hit the next re-index interval, which would suck mightily.
+                self.log.info('Search indexing has worked for an hour; giving up for now.')
+                self.flush()
+                return
+            if now - last_update > 60:
+                self.log.info('Search indexing continues: %d/%d' % (min(count, len(work)), len(work)))
+                last_update = time.time()
+                self.flush()
+        self.log.info('Search index completed.')
+        self.flush()
+

=== modified file 'loggerhead/util.py'
--- loggerhead/util.py	2006-12-15 09:20:04 +0000
+++ loggerhead/util.py	2006-12-20 09:29:42 +0000
@@ -22,7 +22,9 @@
 import logging
 import re
 import sha
+import sys
 import threading
+import traceback
 
 import turbogears
 
@@ -63,7 +65,15 @@
 def ago(timestamp):
     now = datetime.datetime.now()
     return timespan(now - timestamp) + ' ago'
-    
+
+
+def fix_year(year):
+    if year < 70:
+        year += 2000
+    if year < 100:
+        year += 1900
+    return year
+
 
 class Container (object):
     """
@@ -102,6 +112,12 @@
     return text[:limit] + '...'
 
 
+def to_utf8(s):
+    if isinstance(s, unicode):
+        return s.encode('utf-8')
+    return s
+
+
 STANDARD_PATTERN = re.compile(r'^(.*?)\s*<(.*?)>\s*$')
 EMAIL_PATTERN = re.compile(r'[-\w\d\+_!%\.]+@[-\w\d\+_!%\.]+')
 
@@ -159,13 +175,18 @@
             out.insert(0, -n)
 
 
+# only do this if unicode turns out to be a problem
+#_BADCHARS_RE = re.compile(ur'[\u007f-\uffff]')
+
 def html_clean(s):
     """
     clean up a string for html display.  expand any tabs, encode any html
     entities, and replace spaces with '&nbsp;'.  this is primarily for use
     in displaying monospace text.
     """
-    s = cgi.escape(s.expandtabs()).replace(' ', '&nbsp;')
+    s = cgi.escape(s.expandtabs())
+#    s = _BADCHARS_RE.sub(lambda x: '&#%d;' % (ord(x.group(0)),), s)
+    s = s.replace(' ', '&nbsp;')
     return s
 
 
@@ -187,59 +208,106 @@
     return format % value
 
 
+KILO = 1024
+MEG = 1024 * KILO
+GIG = 1024 * MEG
+P95_MEG = int(0.9 * MEG)
+P95_GIG = int(0.9 * GIG)
+
+def human_size(size, min_divisor=0):
+    size = int(size)
+    if (size == 0) and (min_divisor == 0):
+        return '0'
+    if (size < 512) and (min_divisor == 0):
+        return str(size)
+
+    if (size >= P95_GIG) or (min_divisor >= GIG):
+        divisor = GIG
+    elif (size >= P95_MEG) or (min_divisor >= MEG):
+        divisor = MEG
+    else:
+        divisor = KILO
+    
+    dot = size % divisor
+    base = size - dot
+    dot = dot * 10 // divisor
+    base //= divisor
+    if dot >= 10:
+        base += 1
+        dot -= 10
+    
+    out = str(base)
+    if (base < 100) and (dot != 0):
+        out += '.%d' % (dot,)
+    if divisor == KILO:
+        out += 'K'
+    elif divisor == MEG:
+        out += 'M'
+    elif divisor == GIG:
+        out += 'G'
+    return out
+    
+
 def fill_in_navigation(history, navigation):
     """
     given a navigation block (used by the template for the page header), fill
     in useful calculated values.
     """
-    navigation.position = history.get_revid_sequence(navigation.revlist, navigation.revid)
-    navigation.count = len(navigation.revlist)
+    navigation.position = history.get_revid_sequence(navigation.revid_list, navigation.revid)
+    if navigation.position is None:
+        navigation.position = 0
+    navigation.count = len(navigation.revid_list)
     navigation.page_position = navigation.position // navigation.pagesize + 1
-    navigation.page_count = (len(navigation.revlist) + (navigation.pagesize - 1)) // navigation.pagesize
+    navigation.page_count = (len(navigation.revid_list) + (navigation.pagesize - 1)) // navigation.pagesize
     
     def get_offset(offset):
-        if (navigation.position + offset < 0) or (navigation.position > navigation.count - 1):
+        if (navigation.position + offset < 0) or (navigation.position + offset > navigation.count - 1):
             return None
-        return navigation.revlist[navigation.position + offset]
+        return navigation.revid_list[navigation.position + offset]
     
     navigation.prev_page_revid = get_offset(-1 * navigation.pagesize)
     navigation.next_page_revid = get_offset(1 * navigation.pagesize)
+    
+    params = { 'file_id': navigation.file_id }
+    if getattr(navigation, 'query', None) is not None:
+        params['q'] = navigation.query
+    else:
+        params['start_revid'] = navigation.start_revid
+        
     if navigation.prev_page_revid:
-        navigation.prev_page_url = turbogears.url([ navigation.scan_url, navigation.prev_page_revid ], path=navigation.path, start_revid=navigation.start_revid)
+        navigation.prev_page_url = turbogears.url([ navigation.scan_url, navigation.prev_page_revid ], **params)
     if navigation.next_page_revid:
-        navigation.next_page_url = turbogears.url([ navigation.scan_url, navigation.next_page_revid ], path=navigation.path, start_revid=navigation.start_revid)
-
-
-# global branch history & cache
-
-_history = None
-_history_lock = threading.Lock()
-
-def get_history():
-    global _history
-    from loggerhead.history import History
-    
-    config = get_config()
-    
-    _history_lock.acquire()
-    try:
-        if (_history is None) or _history.out_of_date():
-            log.debug('Reload branch history...')
-            if _history is not None:
-                _history.dont_use_cache()
-            _history = History.from_folder(config.get('folder'))
-            _history.use_cache(config.get('cachepath'))
-        return _history
-    finally:
-        _history_lock.release()
-
-
-_config = None
-
-def set_config(config):
-    global _config
-    _config = config
-
-def get_config():
-    return _config
+        navigation.next_page_url = turbogears.url([ navigation.scan_url, navigation.next_page_revid ], **params)
+
+
+def log_exception(log):
+    for line in ''.join(traceback.format_exception(*sys.exc_info())).split('\n'):
+        log.debug(line)
+
+
+def decorator(unbound):
+    def new_decorator(f):
+        g = unbound(f)
+        g.__name__ = f.__name__
+        g.__doc__ = f.__doc__
+        g.__dict__.update(f.__dict__)
+        return g
+    new_decorator.__name__ = unbound.__name__
+    new_decorator.__doc__ = unbound.__doc__
+    new_decorator.__dict__.update(unbound.__dict__)
+    return new_decorator
+
+
+# common threading-lock decorator
+def with_lock(lockname):
+    @decorator
+    def _decorator(unbound):
+        def locked(self, *args, **kw):
+            getattr(self, lockname).acquire()
+            try:
+                return unbound(self, *args, **kw)
+            finally:
+                getattr(self, lockname).release()
+        return locked
+    return _decorator
 

=== modified file 'start-loggerhead.py'
--- start-loggerhead.py	2006-12-15 10:05:25 +0000
+++ start-loggerhead.py	2006-12-20 09:29:42 +0000
@@ -25,4 +25,4 @@
 
 from loggerhead.controllers import Root
 
-turbogears.start_server(Root())
+turbogears.start_server(Root)

